/*******************************************************************************************\


     #######   #####   ######    #####    #####    #####   ######  ######  #     #
        #     #     #  #     #  #     #  #     #  #     #  #       #       ##   ##
        #           #  #     #        #  #     #  #     #  #       #       # # # #
        #      #####   #     #   #####   #     #  #     #  #####   #####   #  #  #
        #           #  #     #  #        #     #  #     #  #       #       #     #
        #     #     #  #     #  #        #     #  #     #  #       #       #     #
        #      #####   ######   #######   #####    #####   #       ######  #     #



                             Converter from T3d to OOFEM


                         Copyright: Daniel Rypl 1998 - 2013



                        Czech Technical University in Prague
                Faculty of Civil Engineering, Department of Mechanics
                      Thakurova 7, 166 29 Prague, Czech Republic
                               email: drypl@fsv.cvut.cz


\*******************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#pragma warning(disable : 4996)  // to avoid warnings for unsafe functions like fopen, etc

#if defined(_WIN32) || defined(_WIN64)
#define strcasecmp _stricmp 
#define strncasecmp _strnicmp
#endif

#define VERSION        "2.4"

/* handle number of initial conditions specified in the record with numbers of oofem input file records */

#define SUPPORT_FOR_NIC


/* caution: this version does not process property specification within boundary entity output
	 on all 2D and 3D elements;
	 this fact should not influence functionality */

/* note: since the ctrl file is processed before t3d output file, all relevant (loaded, supported,
	 ets.) entities are instanciated (even those that are not detected from t3d output because
	 there is no node on it (single element entity) */


#define FILE_NAME_SIZE   128
#define BUFFER_SIZE    65536

#define KEY_SIZE          32


#define SEPARATOR      " \n\r"
#define REMARK           '#'

#define VERTEX_ENTITY      1
#define CURVE_ENTITY       2
#define SURFACE_ENTITY     3
#define REGION_ENTITY      4
#define PATCH_ENTITY       5
#define SHELL_ENTITY       6

#define LINEAR             1
#define QUADRATIC          2

#define TRI_TETRA          3
#define QUAD_HEXA          4

#define GENERAL_ERROR      1
#define LONG_NAME_ERROR    2
#define FILE_OPEN_ERROR    3
#define FILE_READ_ERROR    4
#define FILE_WRITE_ERROR   5
#define MEMORY_ERROR       6


#define DEFAULT_FIRST
/* #define DEFAULT_LAST */


typedef enum logic{
	NO = 0,
	YES = 1
}logic;


typedef struct list_rec list_rec;

struct list_rec{
	void *item;
	list_rec *next;
};


typedef struct entity_rec entity_rec;
typedef struct entities_rec entities_rec;

struct entity_rec{
	long id;
	logic selected;
	list_rec *node_prop_list;
	list_rec *elem_prop_list;
	char *elem_type;
	char *edge_type;
	char *tria_type;
	char *quad_type;
	char *tetra_type;
	char *pyram_type;
	char *wedge_type;
	char *hexa_type;
	list_rec *bload_id_list;
	list_rec *bcode_id_list;
	char *node_type;
	long elems;
	logic on_output;
	list_rec *curve_id;
};

struct entities_rec{
	list_rec *id;
	logic selected;
	list_rec *node_prop_list;
	list_rec *elem_prop_list;
	char *elem_type;
	char *edge_type;
	char *tria_type;
	char *quad_type;
	char *tetra_type;
	char *pyram_type;
	char *wedge_type;
	char *hexa_type;
	list_rec *bload_id_list;
	list_rec *bcode_id_list;
	char *node_type;
};



static list_rec *vertex_list = NULL;
static list_rec *curve_list = NULL;
static list_rec *surface_list = NULL;
static list_rec *patch_list = NULL;
static list_rec *shell_list = NULL;
static list_rec *region_list = NULL;

static list_rec *vertices_list = NULL;
static list_rec *curves_list = NULL;
static list_rec *surfaces_list = NULL;
static list_rec *patches_list = NULL;
static list_rec *shells_list = NULL;
static list_rec *regions_list = NULL;

static list_rec *property_list = NULL;

static list_rec *string_list = NULL;
static list_rec *id_list = NULL;


static logic vertex_select = NO;
static logic curve_select = NO;
static logic surface_select = NO;
static logic patch_select = NO;
static logic shell_select = NO;
static logic region_select = NO;

static logic multi_select = NO;
static logic property_select = NO;
static logic entity_select = NO;
static logic node_elem_select = NO;


static long *node_flag = NULL;
static long *con_node1 = NULL;
static long *con_node2 = NULL;


static enum keyword{
	VERTEX,
	CURVE,
	SURFACE,
	PATCH,
	SHELL,
	REGION,
	ALL,
	EXCEPT,
	PROPERTY,
	NODEPROP,
	ELEMPROP,
	NODETYPE,
	ELEMTYPE,
	EDGETYPE,
	TRIATYPE,
	QUADTYPE,
	TETRATYPE,
	HEXATYPE,
	BLOAD,
	BOUNDARYLOAD,         /* same as bload */
	BCODE,
	BOUNDARYCODE,         /* same as bcode */
	ELEMWEIGHT,
	EDGEWEIGHT,
	TRIAWEIGHT,
	QUADWEIGHT,
	TETRAWEIGHT,
	PYRAMWEIGHT,
	WEDGEWEIGHT,
	HEXAWEIGHT,
	WIDTH,
	PARTITION,
	MASTER,
	CANCEL,
	KEY_NUM
}key, last_key = KEY_NUM;

static char keyword[KEY_NUM][KEY_SIZE] = {
	"VERTEX",
	"CURVE",
	"SURFACE",
	"PATCH",
	"SHELL",
	"REGION",
	"ALL",
	"EXCEPT",
	"PROPERTY",
	"NODEPROP",
	"ELEMPROP",
	"NODETYPE",
	"ELEMTYPE",
	"EDGETYPE",
	"TRIATYPE",
	"QUADTYPE",
	"TETRATYPE",
	"HEXATYPE",
	"BLOAD",
	"BOUNDARYLOAD",
	"BCODE",
	"BOUNDARYCODE",
	"ELEMWEIGHT",
	"EDGEWEIGHT",
	"TRIAWEIGHT",
	"QUADWEIGHT",
	"TETRAWEIGHT",
	"PYRAMWEIGHT",
	"WEDGEWEIGHT",
	"HEXAWEIGHT",
	"WIDTH",
	"PARTITION",
	"MASTER",
	"CANCEL"
};


typedef enum key_case{
	MIXED_CASE,
	UPPER_CASE,
	LOWER_CASE
}key_case;

#define KEY_CASE UPPER_CASE


static char line_buffer[BUFFER_SIZE], temp_buffer[BUFFER_SIZE];

static FILE *active_in_file = NULL, *active_out_file = NULL;


static list_rec *
add_item_to_list_head(list_rec *any_list, void *item);
static list_rec *
deep_destroy_list(list_rec *any_list);
static list_rec *
destroy_list(list_rec *any_list);
static list_rec *
reverse_list(list_rec *any_list);
static long
get_list_size(list_rec *any_list);


static char *
get_next_record(char *err_msg);
static char *
get_next_relevant_record(char *err_msg);

static char *
get_first_token(char *err_msg);
static char *
get_next_token(char *err_msg);
static int
get_token_key(char *token, char keyword[1][KEY_SIZE], int keynum);

long
get_next_int(char *err_msg);
double
get_next_fpn(char *err_msg);

long
get_int_from_token(char *token);
double
get_fpn_from_token(char *token);

static void
write_current_record(char *err_msg);
static void
read_write_next_relevant_record(char *in_err_msg, char *out_err_msg);
static void
read_write_unrelevant_records(char *in_err_msg, char *out_err_msg);
static void
read_write_until_next_relevant_record(char *in_err_msg, char *out_err_msg);


static long *
get_id(list_rec *id_list, long id);
static long *
create_id(long id);

static entity_rec *
get_entity_id(list_rec *entity_list, long id);
static entity_rec *
create_entity_id(long id);
	
static entities_rec *
create_entities(void);

static entity_rec *
create_missing_entity_id(long id, int entity_type);


static void
process_entity(entity_rec *ent, int entity_type);
static void
process_entity_list(list_rec **entity_list, list_rec **entities_list, char *entity_name);
static void
process_entities_list(entities_rec *ents, char *entity_name);

static void
deselect_entity_list(list_rec *entity_list);
static void
deselect_entities_list(list_rec *entities_list);


static void
reverse_node_elem_entity_list(list_rec *entity_list);


static void
apply_node_properties(char *string);
static void
apply_node_properties_entity_list(list_rec *entity_list, char *string);
static void
apply_node_properties_entities_list(list_rec *entities_list, char *string);

static void
apply_elem_properties(char *string);
static void
apply_elem_properties_entity_list(list_rec *entity_list, char *string);
static void
apply_elem_properties_entities_list(list_rec *entities_list, char *string);

static void
apply_node_type(char *string);
static void
apply_node_type_entity_list(list_rec *entity_list, char *string, char *entity_name);
static void
apply_node_type_entities_list(list_rec *entities_list, char *string, char *entity_name);

static void
apply_elem_type(char *string);
static void
apply_elem_type_entity_list(list_rec *entity_list, char *string, char *entity_name);
static void
apply_elem_type_entities_list(list_rec *entities_list, char *string, char *entity_name);

static void
apply_elem_bload(long *ident);
static void
apply_elem_bload_entity_list(list_rec *entity_list, long *ident);
static void
apply_elem_bload_entities_list(list_rec *entities_list, long *ident);

static void
apply_elem_bcode(long *ident);
static void
apply_elem_bcode_entity_list(list_rec *entity_list, long *ident);
static void
apply_elem_bcode_entities_list(list_rec *entities_list, long *ident);

static logic
get_linear_edge_flag(long node1_id, long node2_id, long curve_id);
static logic
get_quadratic_edge_flag(long midnode_id);

static void
delete_entity_properties(list_rec *entity_list);
static void
delete_entities_properties(list_rec *entities_list);
static void
destroy_entity_list(list_rec *entity_list);
static void
destroy_entities_list(list_rec *entities_list);

static void
check_entity_list(list_rec *entity_list, char *entity_name);

static char *
store_string(char *string);

static void
exit_error_message(char *message, int exit_code, int line);



#define error_message(MSG, CODE)    exit_error_message(MSG, CODE, __LINE__)


static logic warning = NO;


int
main(int argc, char **argv)
{
	long nodes = 0, edges = 0, trias = 0, quads = 0, tetras = 0, hexas = 0, wedges = 0, pyrams = 0, elems = 0;
	long node_id, elem_id, entity_id, property_id, last_entity_id = 0, last_property_id = 0;
	long node1_id, node2_id, node3_id, node4_id, node5_id, node6_id, node7_id, node8_id;
	long nd1_id, nd2_id, nd3_id, nd4_id, nd5_id, nd6_id, nd7_id, nd8_id, nd9_id, nd10_id, nd11_id, nd12_id;
	long bound_cur_id[12], bound_cur_prop[12], bound_ent_id[6], bound_ent_tp[6], bound_ent_prop[6];
	long count, load_id, code_id, *ident = NULL;
	logic bflag[6], cflag[12], edge_flag[12], bnd_prop;
	int csect, mater, bcond, nic = 0, timef, records, recs, add_recs, length, num;
	int i, j, mesh_type, elem_degree, renum_type, output_type, entity_type, last_entity_type = 0;
	double x, y, z;
	/*	double tx, ty, tz, nx, ny, nz; */
	FILE *control_file = NULL, *t3d_out_file = NULL;
	FILE *oofem_in_file = NULL, *oofem_domain_file = NULL;
	char control_name[FILE_NAME_SIZE], t3d_out_name[FILE_NAME_SIZE];
	char oofem_in_name[FILE_NAME_SIZE], oofem_domain_name[FILE_NAME_SIZE + 7];
	char *token = NULL, *string = NULL, err_msg[256], in_err_msg[256], out_err_msg[256], *elem_type;
	list_rec *node_list = NULL, *elem_list = NULL, *entity_list = NULL, *load_id_list = NULL, *code_id_list = NULL;
	entity_rec *entity = NULL, *ent = NULL;
	logic write_rec, domain = NO, help = NO;

	int p2h_fc[5] = {1, 3, 4, 5, 6};                                /* t3d pyram -> oofem hexa */
	int w2h_fc[5] = {1, 2, 3, 4, 6};                                /* t3d wedge -> oofem hexa */
	int h2h_fc[6] = {1, 2, 3, 4, 5, 6};                             /* t3d hexa  -> oofem hexa */  /* identity */

	int p2h_ed[8] = {1, 2, 3, 4, 5, 6, 7, 8};                       /* t3d pyram -> oofem hexa */  /* identity */
	int w2h_ed[9] = {1, 2, 4, 9, 10, 12, 5, 6, 8};                  /* t3d wedge -> oofem hexa */
	int h2h_ed[12] = {1, 2, 3, 4, 9, 10, 11, 12, 5, 6, 7, 8};       /* t3d hexa  -> oofem hexa */

	if(argc == 2){
		if(strcmp(*(argv + 1), "-h") == 0)help = YES;
	}

	fprintf(stdout, "\n");
	fprintf(stdout, "**************************************\n");
	fprintf(stdout, "T3d2oofem: Converter from T3d to Oofem\n");
	fprintf(stdout, "Copyright: Daniel Rypl, 2012\n");
	fprintf(stdout, "Version:   %s\n", VERSION);
	fprintf(stdout, "**************************************\n");
	fprintf(stdout, "\n");

	if(argc < 4 || help == YES){
		fprintf(stdout, "Usage: t3d2oofem [-h] control_file t3d_output_file oofem_input_file [-din]\n\n");
		fprintf(stdout, "       -h     prints usage and control file keywords\n");
		fprintf(stdout, "       -din   outputs also domain input file (oofem input file without\n");
		fprintf(stdout, "              certain control records) needed for adaptive analysis\n\n");

		if(help == YES){
			fprintf(stdout, "\n");
			fprintf(stdout, "Control file keywords\n");
			fprintf(stdout, "=====================\n");
			fprintf(stdout, "\n");
			fprintf(stdout, "Model entity specification\n");
			fprintf(stdout, "--------------------------\n");
			fprintf(stdout, "{Vertex | Curve | Surface | Patch | Shell | Region} {All [Except <int>] | <int>}\n");
			fprintf(stdout, "\n");
			fprintf(stdout, "Property entity specification\n");
			fprintf(stdout, "-----------------------------\n");
			fprintf(stdout, "Property {All [Except <int>] | <int>}\n");
			fprintf(stdout, "\n");
			fprintf(stdout, "Model and property specification cancelation\n");
			fprintf(stdout, "--------------------------------------------\n");
			fprintf(stdout, "Cancel\n");
			fprintf(stdout, "\n");
			fprintf(stdout, "Feature specification\n");
			fprintf(stdout, "---------------------\n");
			fprintf(stdout, "{Nodeprop | Elemprop} string\n");
			fprintf(stdout, "{Edgetype | Triatype | Quadtype | Tetratype | Hexatype} string\n");
			fprintf(stdout, "{Bload | Boundaryload | Bcode | Boundarycode} int\n");
			fprintf(stdout, "\n");
			return(0);
		}

		return(1);
	}
	
	if(strlen(argv[1]) >= FILE_NAME_SIZE)error_message("Too long control_file name", LONG_NAME_ERROR);
	if(strlen(argv[2]) >= FILE_NAME_SIZE)error_message("Too long t3d_output_file name", LONG_NAME_ERROR);
	if(strlen(argv[3]) >= FILE_NAME_SIZE)error_message("Too long oofem_input_file name", LONG_NAME_ERROR);
	
	strncpy(control_name, argv[1], FILE_NAME_SIZE);
	strncpy(t3d_out_name, argv[2], FILE_NAME_SIZE);
	strncpy(oofem_in_name, argv[3], FILE_NAME_SIZE);
	
	argc -= 4;
	argv += 4;
	while(argc--){
		if(strcmp(*argv, "-din") == 0){
			domain = YES;
			argv++;
			continue;
		}
		fprintf(stderr, "Unknown option: %s\n", *argv);
	}


	if((control_file = fopen(control_name, "r")) == NULL){
		sprintf(err_msg, "File %s opening error", control_name);
		error_message(err_msg, FILE_OPEN_ERROR);
	}
	if((t3d_out_file = fopen(t3d_out_name, "r")) == NULL){
		sprintf(err_msg, "File %s opening error", t3d_out_name);
		error_message(err_msg, FILE_OPEN_ERROR);
	}
	if((oofem_in_file = fopen(oofem_in_name, "w")) == NULL){
		sprintf(err_msg, "File %s opening error", oofem_in_name);
		error_message(err_msg, FILE_OPEN_ERROR);
	}

	active_in_file = control_file;
	active_out_file = oofem_in_file;

	sprintf(in_err_msg, "Unexpected end of control file %s", control_name);
	sprintf(out_err_msg, "File %s writing error", oofem_in_name);
	
	read_write_until_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM output file record */
	read_write_until_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM problem description record */
	read_write_until_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM analysis type record */

	add_recs = 0;                                                    /* OOFEM analysis additional records */
	length = (int)strlen(string = line_buffer);
	for(i = 0; i < length; i++, string++){
		if(strncasecmp(string, "nmsteps", 7) == 0){                    /* OOFEM metastep record */
			string += 7;
			sscanf(string, "%d", &recs);
			add_recs += recs;
		}
		if(strncasecmp(string, "nmodules", 8) == 0){                   /* OOFEM export module record */
			string += 8;
			sscanf(string, "%d", &recs);
			add_recs += recs;
		}
		if(strncasecmp(string, "ninitmodules", 12) == 0){              /* OOFEM init module record */
			string += 12;
			sscanf(string, "%d", &recs);
			add_recs += recs;
		}
		if(strncasecmp(string, "nxfemman", 8) == 0){                   /* OOFEM xfem manager record */
			string += 8;
			sscanf(string, "%d", &recs);
			add_recs += recs;
		}
	}

	for(i = 0; i < add_recs; i++)
		read_write_until_next_relevant_record(in_err_msg, out_err_msg);/* OOFEM analysis additional records */

	read_write_until_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM domain record */
	read_write_until_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM output manager record */

	read_write_unrelevant_records(in_err_msg, out_err_msg);

#ifdef SUPPORT_FOR_NIC
	if(sscanf(line_buffer, "%d %d %d %d %d", &csect, &mater, &bcond, &nic, &timef) != 5){
#else
	if(sscanf(line_buffer, "%d %d %d %d", &csect, &mater, &bcond, &timef) != 4){
#endif
		sprintf(err_msg, "File %s reading error", control_name);
		error_message(err_msg, FILE_READ_ERROR);
	}

/* skip node, element, cross section, material, boundary condition and time function records */
	
	records = csect + mater + bcond + nic + timef;
	while(records--){
		get_next_relevant_record(in_err_msg);
	}

	while((token = get_first_token(NULL)) != NULL){
		key = get_token_key(token, keyword, KEY_NUM);

		switch(key){
		case VERTEX:
		case CURVE:
		case SURFACE:
		case PATCH:
		case SHELL:
		case REGION:
			if(property_select == YES && node_elem_select == NO){
				sprintf(err_msg, "Keyword %s not expected together with any entity keyword", keyword[PROPERTY]);
				error_message(err_msg, GENERAL_ERROR);
			}
			if(node_elem_select == YES){
				if(entity_select == YES){
					if(vertex_select == YES)deselect_entity_list(vertex_list);
					if(curve_select == YES)deselect_entity_list(curve_list);
					if(surface_select == YES)deselect_entity_list(surface_list);
					if(patch_select == YES)deselect_entity_list(patch_list);
					if(shell_select == YES)deselect_entity_list(shell_list);
					if(region_select == YES)deselect_entity_list(region_list);
					
					if(multi_select == YES){
						if(vertex_select == YES)deselect_entities_list(vertices_list);
						if(curve_select == YES)deselect_entities_list(curves_list);
						if(surface_select == YES)deselect_entities_list(surfaces_list);
						if(patch_select == YES)deselect_entities_list(patches_list);
						if(shell_select == YES)deselect_entities_list(shells_list);
						if(region_select == YES)deselect_entities_list(regions_list);
					}
					
					vertex_select = curve_select = surface_select = patch_select = shell_select = region_select = NO;
				}
				if(property_select == YES)deselect_entity_list(property_list);
				
				multi_select = entity_select = property_select = node_elem_select = NO;
			}
			entity_select = YES;
			break;
			
		case PROPERTY:
			if(entity_select == YES && node_elem_select == NO){
				sprintf(err_msg, "Keyword %s not expected together with any entity keyword", keyword[PROPERTY]);
				error_message(err_msg, GENERAL_ERROR);
			}
			if(node_elem_select == YES){
				if(entity_select == YES){
					if(vertex_select == YES)deselect_entity_list(vertex_list);
					if(curve_select == YES)deselect_entity_list(curve_list);
					if(surface_select == YES)deselect_entity_list(surface_list);
					if(patch_select == YES)deselect_entity_list(patch_list);
					if(shell_select == YES)deselect_entity_list(shell_list);
					if(region_select == YES)deselect_entity_list(region_list);
					
					if(multi_select == YES){
						if(vertex_select == YES)deselect_entities_list(vertices_list);
						if(curve_select == YES)deselect_entities_list(curves_list);
						if(surface_select == YES)deselect_entities_list(surfaces_list);
						if(patch_select == YES)deselect_entities_list(patches_list);
						if(shell_select == YES)deselect_entities_list(shells_list);
						if(region_select == YES)deselect_entities_list(regions_list);
					}

					vertex_select = curve_select = surface_select = patch_select = shell_select = region_select = NO;
				}
				if(property_select == YES)deselect_entity_list(property_list);

				multi_select = entity_select = property_select = node_elem_select = NO;
			}
			break;

		case NODEPROP:
		case ELEMPROP:
			if(entity_select == NO && property_select == NO)
				error_message("Entity or property specification missing", GENERAL_ERROR);

			if((token = get_next_token(NULL)) == NULL)
				error_message("Missing node or element property", GENERAL_ERROR);

/* do not use line_buffer (memory overlapping may occur) */

			*temp_buffer = '\0';
			while(token != NULL){

/* each token must be preceeded by space */

				strcat(temp_buffer, " ");
				strcat(temp_buffer, token);
				token = get_next_token(NULL);
			}
			string = store_string(temp_buffer);
			string_list = add_item_to_list_head(string_list, (void *)string);
			
			node_elem_select = YES;
			break;

		case NODETYPE:
			if(entity_select == NO)
				error_message("Entity specification missing", GENERAL_ERROR);

			if((token = get_next_token(NULL)) == NULL)
				error_message("Missing node type", GENERAL_ERROR);

			string = store_string(token);
			string_list = add_item_to_list_head(string_list, (void *)string);

			if((token = get_next_token(NULL)) != NULL)
				error_message("Redundant node type specification detected", GENERAL_ERROR);
			
			node_elem_select = YES;
			break;

		case ELEMTYPE:      /* for compatibility reasons */
		case EDGETYPE:
		case TRIATYPE:
		case QUADTYPE:
		case TETRATYPE:
		case HEXATYPE:
			if(entity_select == NO)
				error_message("Entity specification missing", GENERAL_ERROR);

			if((token = get_next_token(NULL)) == NULL)
				error_message("Missing element type", GENERAL_ERROR);

			string = store_string(token);
			string_list = add_item_to_list_head(string_list, (void *)string);

			if((token = get_next_token(NULL)) != NULL)
				error_message("Redundant element type specification detected", GENERAL_ERROR);
			
			node_elem_select = YES;
			break;

		case BLOAD:
		case BOUNDARYLOAD:
		case BCODE:
		case BOUNDARYCODE:
			if(entity_select == NO)
				error_message("Entity specification missing", GENERAL_ERROR);
			
			node_elem_select = YES;
			break;

		case ELEMWEIGHT:
		case EDGEWEIGHT:
		case TRIAWEIGHT:
		case QUADWEIGHT:
		case TETRAWEIGHT:
		case PYRAMWEIGHT:
		case WEDGEWEIGHT:
		case HEXAWEIGHT:
		case WIDTH:
		case PARTITION:
		case MASTER:
		case CANCEL:
			node_elem_select = YES;
			break;

		default:
			sprintf(err_msg, "Unexpected or invalid keyword \"%s\" encountered", token);
			error_message(err_msg, GENERAL_ERROR);
		}
		
		switch(key){
		case VERTEX:
			process_entity_list(&vertex_list, &vertices_list, keyword[VERTEX]);
			vertex_select = YES;
			break;
		case CURVE:
			process_entity_list(&curve_list, &curves_list, keyword[CURVE]);
			curve_select = YES;
			break;
		case SURFACE:
			process_entity_list(&surface_list, &surfaces_list, keyword[SURFACE]);
			surface_select = YES;
			break;
		case PATCH:
			process_entity_list(&patch_list, &patches_list, keyword[PATCH]);
			patch_select = YES;
			break;
		case SHELL:
			process_entity_list(&shell_list, &shells_list, keyword[SHELL]);
			shell_select = YES;
			break;
		case REGION:
			process_entity_list(&region_list, &regions_list, keyword[REGION]);
			region_select = YES;
			break;
			
		case PROPERTY:
			process_entity_list(&property_list, NULL, keyword[PROPERTY]);
			property_select = YES;
			break;
			
		case NODEPROP:
			apply_node_properties(string);
			break;
		case ELEMPROP:
			apply_elem_properties(string);
			break;

		case NODETYPE:
			apply_node_type(string);
			break;

		case ELEMTYPE:
		case EDGETYPE:
		case TRIATYPE:
		case QUADTYPE:
		case TETRATYPE:
		case HEXATYPE:
			apply_elem_type(string);
			break;
			
		case BLOAD:
		case BOUNDARYLOAD:
			load_id = 0;
			while((token = get_next_token(NULL)) != NULL){
				if((load_id = get_int_from_token(token)) <= 0){
					sprintf(err_msg, "Invalid boundary load id (%ld)", load_id);
					error_message(err_msg, GENERAL_ERROR);
				}
				if((ident = get_id(id_list, load_id)) == NULL){
					ident = create_id(load_id);
					id_list = add_item_to_list_head(id_list, (void *)ident);
				}
				apply_elem_bload(ident);
			}
			if(load_id == 0){
				sprintf(err_msg, "Missing boundary load id");
				error_message(err_msg, GENERAL_ERROR);
			}
			break;
			
		case BCODE:
		case BOUNDARYCODE:
			code_id = 0;
			while((token = get_next_token(NULL)) != NULL){
				if((code_id = get_int_from_token(token)) <= 0){
					sprintf(err_msg, "Invalid boundary code id (%ld)", code_id);
					error_message(err_msg, GENERAL_ERROR);
				}
				if((ident = get_id(id_list, code_id)) == NULL){
					ident = create_id(code_id);
					id_list = add_item_to_list_head(id_list, (void *)ident);
				}
				apply_elem_bcode(ident);
			}
			if(code_id == 0){
				sprintf(err_msg, "Missing boundary code id");
				error_message(err_msg, GENERAL_ERROR);
			}
			break;

		case ELEMWEIGHT:
		case EDGEWEIGHT:
		case TRIAWEIGHT:
		case QUADWEIGHT:
		case TETRAWEIGHT:
		case PYRAMWEIGHT:
		case WEDGEWEIGHT:
		case HEXAWEIGHT:
		case WIDTH:
		case PARTITION:
		case MASTER:
		case CANCEL:
			break;
			
		default:
			sprintf(err_msg, "Unexpected or invalid keyword \"%s\" encountered", token);
			error_message(err_msg, GENERAL_ERROR);
		}
	}
	
	if(node_elem_select == NO){
		if(entity_select == YES || property_select == YES)
			error_message("Node or elem properties missing", GENERAL_ERROR);
	}
	
	if(entity_select == YES){
		if(vertex_select == YES)deselect_entity_list(vertex_list);
		if(curve_select == YES)deselect_entity_list(curve_list);
		if(surface_select == YES)deselect_entity_list(surface_list);
		if(patch_select == YES)deselect_entity_list(patch_list);
		if(shell_select == YES)deselect_entity_list(shell_list);
		if(region_select == YES)deselect_entity_list(region_list);
		
		if(multi_select == YES){
			if(vertex_select == YES)deselect_entities_list(vertices_list);
			if(curve_select == YES)deselect_entities_list(curves_list);
			if(surface_select == YES)deselect_entities_list(surfaces_list);
			if(patch_select == YES)deselect_entities_list(patches_list);
			if(shell_select == YES)deselect_entities_list(shells_list);
			if(region_select == YES)deselect_entities_list(regions_list);
		}
		
		vertex_select = curve_select = surface_select = patch_select = shell_select = region_select = NO;
	}
	if(property_select == YES)deselect_entity_list(property_list);
	
	multi_select = entity_select = property_select = node_elem_select = NO;

/* it is better to reverse the list in the instance of use
	 (not all entities may be actually used) */
/*
#ifdef DEFAULT_FIRST
	reverse_node_elem_entity_list(vertex_list);
	reverse_node_elem_entity_list(curve_list);
	reverse_node_elem_entity_list(surface_list);
	reverse_node_elem_entity_list(patch_list);
	reverse_node_elem_entity_list(shell_list);
	reverse_node_elem_entity_list(region_list);
#endif
*/
	reverse_node_elem_entity_list(property_list);
	
	active_in_file = t3d_out_file;
	sprintf(in_err_msg, "File %s reading error", t3d_out_name);
	
	get_next_relevant_record(in_err_msg);
	if(sscanf(line_buffer, "%d %d %d %d", &mesh_type, &elem_degree, &renum_type, &output_type) != 4)
		error_message(in_err_msg, FILE_READ_ERROR);

	if((output_type & 8) == 0)
		error_message("Invalid t3d -p option (8 or 512 missing)", GENERAL_ERROR);

	get_next_relevant_record(in_err_msg);
	switch(mesh_type){
	case TRI_TETRA:
		if(sscanf(line_buffer, "%ld %ld %ld %ld", &nodes, &edges, &trias, &tetras) != 4)
			error_message(in_err_msg, FILE_READ_ERROR);
		elems = edges + trias + tetras;
		break;
	case QUAD_HEXA:
		if(sscanf(line_buffer, "%ld %ld %ld %ld", &nodes, &edges, &quads, &hexas) != 4)
			error_message(in_err_msg, FILE_READ_ERROR);
		elems = edges + quads + hexas;
		break;
	case TRI_TETRA + QUAD_HEXA:
		if(sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld", &nodes, &edges, &trias, &quads, &tetras, &pyrams, &wedges, &hexas) != 8)
			error_message(in_err_msg, FILE_READ_ERROR);
		elems = edges + trias + quads + tetras + pyrams + wedges + hexas;
		break;
	default:
		error_message("Unknown element type", GENERAL_ERROR);
		break;
	}

	if((output_type & 512) == 0){
		if(tetras != 0 || pyrams != 0 || wedges != 0 || hexas != 0){
			fprintf(stderr, "\n");
			fprintf(stderr, "Warning: Edgeload (if any) applied to 3D elements is ignored\n");
			fprintf(stderr, "         unless t3d was executed with -p 512 option\n\n");
		}
	}

	switch(elem_degree){
	case LINEAR:
	case QUADRATIC:
		break;
	default:
		error_message("Unsupported element degree", GENERAL_ERROR);
	}
	
	if(nodes == 0)
		error_message("Not enough nodes found", GENERAL_ERROR);

	if(elem_degree == LINEAR){
		if((con_node1 = (long *)calloc(nodes, sizeof(long))) == NULL)
			error_message("Memory allocation error", MEMORY_ERROR);
		if((con_node2 = (long *)calloc(nodes, sizeof(long))) == NULL)
			error_message("Memory allocation error", MEMORY_ERROR);
	}
	else{
		if((node_flag = (long *)calloc(nodes, sizeof(long))) == NULL)
			error_message("Memory allocation error", MEMORY_ERROR);
	}

	if(fprintf(oofem_in_file, "ndofman %ld nelem %ld ncrosssect %d nmat %d nbc %d nic %d nltf %d\n", 
						 nodes, elems, csect, mater, bcond, nic, timef) < 0)
		error_message(out_err_msg, FILE_WRITE_ERROR);         /* OOFEM component size record */
	
/* I am using fscanf because I am reading fixed format any item of which is used as
	 flow control variable;
	 to enable use of variable number of parameters on the input line the rest of the record 
	 (at least the new-line character at the end of the record) is read into string 
	 from which the remaining variables are read */

	last_entity_type = 0;
	last_entity_id = 0;
	last_property_id = 0;
	for(i = 0; i < nodes; i++){
		if(fscanf(t3d_out_file, "%ld %lf %lf %lf %d %ld %ld", &node_id, &x, &y, &z, 
							&entity_type, &entity_id, &property_id) != 7)
			error_message(in_err_msg, FILE_READ_ERROR);
		
		get_next_record(in_err_msg);
		switch(entity_type){
		case VERTEX_ENTITY:
			entity_list = vertex_list;
			if(elem_degree == LINEAR)con_node1[node_id - 1] = con_node2[node_id - 1] = -entity_id;
			break;
		case CURVE_ENTITY:
			entity_list = curve_list;
/*			if(sscanf(line_buffer, "%lf %lf %lf", &tx, &ty, &tz) != 3)
				error_message(in_err_msg, FILE_READ_ERROR); */
			break;
		case SURFACE_ENTITY:
			entity_list = surface_list;
/*			if(sscanf(line_buffer, "%lf %lf %lf", &nx, &ny, &nz) != 3)
				error_message(in_err_msg, FILE_READ_ERROR); */
			break;
		case PATCH_ENTITY:
			entity_list = patch_list;
/*			if(sscanf(line_buffer, "%lf %lf %lf", &nx, &ny, &nz) != 3)
				error_message(in_err_msg, FILE_READ_ERROR); */
			break;
		case SHELL_ENTITY:
			entity_list = shell_list;
/*			if(sscanf(line_buffer, "%lf %lf %lf", &nx, &ny, &nz) != 3)
				error_message(in_err_msg, FILE_READ_ERROR); */
			break;
		case REGION_ENTITY:
			entity_list = region_list;
			break;
		default:
			error_message("Unknown entity type", GENERAL_ERROR);
		}
		
		if(last_entity_type != entity_type || last_entity_id != entity_id){
			entity = get_entity_id(entity_list, entity_id);
			if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
			if(entity -> selected == NO){
				
#ifdef DEFAULT_FIRST
				entity -> node_prop_list = reverse_list(entity -> node_prop_list);
				if(entity_type != VERTEX_ENTITY)
					entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif
				
				process_entity(entity, entity_type);
			}
			last_entity_type = entity_type;
			last_entity_id = entity_id;
			entity -> on_output = YES;
		}

		if(entity -> node_type == NULL){
			if(fprintf(oofem_in_file, "node %ld coords 3 %e %e %e", node_id, x, y, z) < 0)
				error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM node record */
		}
		else{
			if(fprintf(oofem_in_file, "%s %ld coords 3 %e %e %e", entity -> node_type, node_id, x, y, z) < 0)
				error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM node record */
		}

		node_list = entity -> node_prop_list;
		while(node_list != NULL){
			string = (char *)(node_list -> item);
			node_list = node_list -> next;
			
			if(fprintf(oofem_in_file, "%s", string) < 0)
				error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM node record */
		}
		
		if(property_id > 0){
			if(last_property_id != property_id){
				ent = get_entity_id(property_list, property_id);
				last_property_id = property_id;
			}
			if(ent != NULL){
				node_list = ent -> node_prop_list;
				
				while(node_list != NULL){
					string = (char *)(node_list -> item);
					node_list = node_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM node record */
				}
			}
		}
		if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
	}
	
	if(elem_degree == LINEAR){
		if(edges != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < edges; i++){
				get_next_relevant_record(in_err_msg);
				if(sscanf(line_buffer, "%ld %ld %ld %d %ld %ld", 
									&elem_id, &node1_id, &node2_id, &entity_type, &entity_id, &property_id) != 6)
					error_message(in_err_msg, FILE_READ_ERROR);
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != CURVE_ENTITY)
						error_message("Invalid edge entity detected", GENERAL_ERROR);
					
					entity = get_entity_id(curve_list, entity_id);
					
/* the following should not be relevant */

					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){
						
#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif

						process_entity(entity, entity_type);
					}

					elem_type = entity -> edge_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
 					if(elem_type == NULL){
						sprintf(err_msg, "Element type specification for edges on %s %ld missing", keyword[CURVE], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
				
				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 2 %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}

				if((load_id_list = entity -> bload_id_list) != NULL){
					count = get_list_size(load_id_list);

					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					while(load_id_list != NULL){
						if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), 1) < 0)
							error_message(out_err_msg, FILE_WRITE_ERROR);      /* OOFEM elem record */
						load_id_list = load_id_list -> next;
					}
				}

				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}
		
		if(trias != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < trias; i++){
				get_next_relevant_record(in_err_msg);
				num = sscanf(line_buffer, "%ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld", 
										 &elem_id, &node1_id, &node2_id, &node3_id, &entity_type, &entity_id, &property_id,
										 &bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], 
										 &bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2]);
				if(num != 10 && num != 13)
					error_message(in_err_msg, FILE_READ_ERROR);
				bnd_prop = (num == 13) ? YES : NO;
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != SURFACE_ENTITY && entity_type != PATCH_ENTITY && entity_type != SHELL_ENTITY)
						error_message("Invalid triangle entity detected", GENERAL_ERROR);
					
					if(entity_type == SURFACE_ENTITY)entity = get_entity_id(surface_list, entity_id);
					if(entity_type == PATCH_ENTITY)entity = get_entity_id(patch_list, entity_id);
					if(entity_type == SHELL_ENTITY)entity = get_entity_id(shell_list, entity_id);

/* the following should not be relevant */

					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){
						
#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif
						
						process_entity(entity, entity_type);
					}

					elem_type = entity -> tria_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
					if(elem_type == NULL){
						if(entity_type == SURFACE_ENTITY)key = SURFACE;
						if(entity_type == PATCH_ENTITY)key = PATCH;
						if(entity_type == SHELL_ENTITY)key = SHELL;
						sprintf(err_msg, "Element type specification for trias on %s %ld missing", keyword[key], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
				
				/* check whether bload and bcode has been processed on particular edge */

				edge_flag[0] = get_linear_edge_flag(node1_id, node2_id, bound_cur_id[0]);
				edge_flag[1] = get_linear_edge_flag(node2_id, node3_id, bound_cur_id[1]);
				edge_flag[2] = get_linear_edge_flag(node3_id, node1_id, bound_cur_id[2]);

				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 3 %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 3; j++){
					bflag[j] = NO;
					if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
						bflag[j] = YES;
						if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 3; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}
				}

				count = 0;
				for(j = 0; j < 3; j++){
					bflag[j] = NO;
					if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
						bflag[j] = YES;
						if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 3; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 3; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}
				
				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}
		
		if(quads != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < quads; i++){
				get_next_relevant_record(in_err_msg);
				num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
										 &elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &entity_type, &entity_id, &property_id,
										 &bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], &bound_cur_id[3],
										 &bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2], &bound_cur_prop[3]);
				if(num != 12 && num != 16)
					error_message(in_err_msg, FILE_READ_ERROR);
				bnd_prop = (num == 16) ? YES : NO;
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != SURFACE_ENTITY && entity_type != PATCH_ENTITY && entity_type != SHELL_ENTITY)
						error_message("Invalid quadrilateral entity detected", GENERAL_ERROR);
					
					if(entity_type == SURFACE_ENTITY)entity = get_entity_id(surface_list, entity_id);
					if(entity_type == PATCH_ENTITY)entity = get_entity_id(patch_list, entity_id);
					if(entity_type == SHELL_ENTITY)entity = get_entity_id(shell_list, entity_id);

/* the following should not be relevant */

					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){
						
#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif
						
						process_entity(entity, entity_type);
					}

					elem_type = entity -> quad_type;
					
					if(elem_type == NULL){
						if(entity_type == SURFACE_ENTITY)key = SURFACE;
						if(entity_type == PATCH_ENTITY)key = PATCH;
						if(entity_type == SHELL_ENTITY)key = SHELL;
						sprintf(err_msg, "Element type specification for quads on %s %ld missing", keyword[key], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
				
				/* check whether bload and bcode has been processed on particular edge */

				edge_flag[0] = get_linear_edge_flag(node1_id, node2_id, bound_cur_id[0]);
				edge_flag[1] = get_linear_edge_flag(node2_id, node3_id, bound_cur_id[1]);
				edge_flag[2] = get_linear_edge_flag(node3_id, node4_id, bound_cur_id[2]);
				edge_flag[3] = get_linear_edge_flag(node4_id, node1_id, bound_cur_id[3]);

				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 4 %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, node4_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 4; j++){
					bflag[j] = NO;
					if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
						bflag[j] = YES;
						if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}
				}

				count = 0;
				for(j = 0; j < 4; j++){
					bflag[j] = NO;
					if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
						bflag[j] = YES;
						if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}
				
				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}

		if(tetras != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < tetras; i++){
				get_next_relevant_record(in_err_msg);
				if((output_type & 512) == 0){
					num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
											 &elem_id, &node1_id, &node2_id, &node3_id, &node4_id, 
											 &entity_type, &entity_id, &property_id,
											 &bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], 
											 &bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3],
											 &bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3]);
					if(num != 16 && num != 20)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = (num == 20) ? YES : NO;
				}
				else{
					if(sscanf(line_buffer, "%ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
										&elem_id, &node1_id, &node2_id, &node3_id, &node4_id, 
										&entity_type, &entity_id, &property_id,
										&bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], 
										&bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3],
										&bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], 
										&bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], &bound_cur_id[3], &bound_cur_id[4], &bound_cur_id[5], 
										&bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2], 
										&bound_cur_prop[3], &bound_cur_prop[4], &bound_cur_prop[5]) != 32)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = YES;
				}
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != REGION_ENTITY)
						error_message("Invalid tetrahedron entity detected", GENERAL_ERROR);
					
					entity = get_entity_id(region_list, entity_id);
					
/* the following should not be relevant */
					
					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){

#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif

						process_entity(entity, entity_type);
					}

					elem_type = entity -> tetra_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
					if(elem_type == NULL){
						sprintf(err_msg, "Element type specification for tetras in %s %ld missing", keyword[REGION], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
					
				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 4 %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, node4_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 4; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}

				if((output_type & 512) != 0){

					/* check whether bload has been processed on particular edge */
					
					edge_flag[0] = get_linear_edge_flag(node1_id, node2_id, bound_cur_id[0]);
					edge_flag[1] = get_linear_edge_flag(node2_id, node3_id, bound_cur_id[1]);
					edge_flag[2] = get_linear_edge_flag(node3_id, node1_id, bound_cur_id[2]);
					edge_flag[3] = get_linear_edge_flag(node1_id, node4_id, bound_cur_id[3]);
					edge_flag[4] = get_linear_edge_flag(node2_id, node4_id, bound_cur_id[4]);
					edge_flag[5] = get_linear_edge_flag(node3_id, node4_id, bound_cur_id[5]);

					for(j = 0; j < 6; j++){
						cflag[j] = NO;
						if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
							cflag[j] = YES;
							if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
								if(ent -> bload_id_list != NULL)
									count += get_list_size(ent -> bload_id_list);
								else
									cflag[j] = NO;
							}
							else
								cflag[j] = NO;
						}
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}

					if((output_type & 512) != 0){
						for(j = 0; j < 6; j++){
							if(cflag[j] == NO)continue;
							ent = get_entity_id(curve_list, bound_cur_id[j]);
							load_id_list = ent -> bload_id_list;
							while(load_id_list != NULL){
								if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)
									error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
								load_id_list = load_id_list -> next;
							}
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 4; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}

				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}

				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}

		if(pyrams != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < pyrams; i++){
				get_next_relevant_record(in_err_msg);
				if((output_type & 512) == 0){
					num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
											 &elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, 
											 &entity_type, &entity_id, &property_id,
											 &bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], 
											 &bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4],
											 &bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4]);
					if(num != 19 && num != 24)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = (num == 24) ? YES : NO;
				}
				else{
					if(sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
										&elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, 
										&entity_type, &entity_id, &property_id,
										&bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], 
										&bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4],
										&bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4],
										&bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], &bound_cur_id[3],
										&bound_cur_id[4], &bound_cur_id[5], &bound_cur_id[6], &bound_cur_id[7],
										&bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2], &bound_cur_prop[3],
										&bound_cur_prop[4], &bound_cur_prop[5], &bound_cur_prop[6], &bound_cur_prop[7]) != 40)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = YES;
				}
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != REGION_ENTITY)
						error_message("Invalid pyramid entity detected", GENERAL_ERROR);
					
					entity = get_entity_id(region_list, entity_id);
					
/* the following should not be relevant */
					
					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){

#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif

						process_entity(entity, entity_type);
					}

					elem_type = entity -> pyram_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
					if(elem_type == NULL){
						sprintf(err_msg, "Element type specification for pyrams in %s %ld missing", keyword[REGION], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
					
				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 8 %ld %ld %ld %ld %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, node4_id, node5_id, node5_id, node5_id, node5_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 5; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}

				if((output_type & 512) != 0){

					/* check whether bload has been processed on particular edge */
					
					edge_flag[0] = get_linear_edge_flag(node1_id, node2_id, bound_cur_id[0]);
					edge_flag[1] = get_linear_edge_flag(node2_id, node3_id, bound_cur_id[1]);
					edge_flag[2] = get_linear_edge_flag(node3_id, node4_id, bound_cur_id[2]);
					edge_flag[3] = get_linear_edge_flag(node4_id, node1_id, bound_cur_id[3]);
					edge_flag[4] = get_linear_edge_flag(node1_id, node5_id, bound_cur_id[4]);
					edge_flag[5] = get_linear_edge_flag(node2_id, node5_id, bound_cur_id[5]);
					edge_flag[6] = get_linear_edge_flag(node3_id, node5_id, bound_cur_id[6]);
					edge_flag[7] = get_linear_edge_flag(node4_id, node5_id, bound_cur_id[7]);

					for(j = 0; j < 8; j++){
						cflag[j] = NO;
						if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
							cflag[j] = YES;
							if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
								if(ent -> bload_id_list != NULL)
									count += get_list_size(ent -> bload_id_list);
								else
									cflag[j] = NO;
							}
							else
								cflag[j] = NO;
						}
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), p2h_fc[j]) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}

					if((output_type & 512) != 0){
						for(j = 0; j < 8; j++){
							if(cflag[j] == NO)continue;
							ent = get_entity_id(curve_list, bound_cur_id[j]);
							load_id_list = ent -> bload_id_list;
							while(load_id_list != NULL){
								if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)   /* p2h_ed[j] */
									error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
								load_id_list = load_id_list -> next;
							}
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 5; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", p2h_fc[j]) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}

				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}

		if(wedges != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < wedges; i++){
				get_next_relevant_record(in_err_msg);
				if((output_type & 512) == 0){
					num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
											 &elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, &node6_id, 
											 &entity_type, &entity_id, &property_id,
											 &bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], 
											 &bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4],
											 &bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4]);
					if(num != 20 && num != 25)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = (num == 25) ? YES : NO;
				}
				else{
					if(sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
										&elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, &node6_id, 
										&entity_type, &entity_id, &property_id,
										&bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], 
										&bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4],
										&bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4],
										&bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], &bound_cur_id[3], &bound_cur_id[4], 
										&bound_cur_id[5], &bound_cur_id[6], &bound_cur_id[7], &bound_cur_id[8],
										&bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2], &bound_cur_prop[3], &bound_cur_prop[4], 
										&bound_cur_prop[5], &bound_cur_prop[6], &bound_cur_prop[7], &bound_cur_prop[8]) != 43)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = YES;
				}
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != REGION_ENTITY)
						error_message("Invalid wedge entity detected", GENERAL_ERROR);
					
					entity = get_entity_id(region_list, entity_id);
					
/* the following should not be relevant */
					
					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){

#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif

						process_entity(entity, entity_type);
					}

					elem_type = entity -> wedge_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
					if(elem_type == NULL){
						sprintf(err_msg, "Element type specification for wedges in %s %ld missing", keyword[REGION], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
					
				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 8 %ld %ld %ld %ld %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, node3_id, node4_id, node5_id, node6_id, node6_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 5; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}

				if((output_type & 512) != 0){

					/* check whether bload has been processed on particular edge */

					edge_flag[0] = get_linear_edge_flag(node1_id, node2_id, bound_cur_id[0]);
					edge_flag[1] = get_linear_edge_flag(node2_id, node3_id, bound_cur_id[1]);
					edge_flag[2] = get_linear_edge_flag(node3_id, node1_id, bound_cur_id[2]);
					edge_flag[3] = get_linear_edge_flag(node4_id, node5_id, bound_cur_id[3]);
					edge_flag[4] = get_linear_edge_flag(node5_id, node6_id, bound_cur_id[4]);
					edge_flag[5] = get_linear_edge_flag(node6_id, node4_id, bound_cur_id[5]);
					edge_flag[6] = get_linear_edge_flag(node1_id, node4_id, bound_cur_id[6]);
					edge_flag[7] = get_linear_edge_flag(node2_id, node5_id, bound_cur_id[7]);
					edge_flag[8] = get_linear_edge_flag(node3_id, node6_id, bound_cur_id[8]);

					for(j = 0; j < 9; j++){
						cflag[j] = NO;
						if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
							cflag[j] = YES;
							if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
								if(ent -> bload_id_list != NULL)
									count += get_list_size(ent -> bload_id_list);
								else
									cflag[j] = NO;
							}
							else
								cflag[j] = NO;
						}
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), w2h_fc[j]) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}

					if((output_type & 512) != 0){
						for(j = 0; j < 9; j++){
							if(cflag[j] == NO)continue;
							ent = get_entity_id(curve_list, bound_cur_id[j]);
							load_id_list = ent -> bload_id_list;
							while(load_id_list != NULL){
								if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), w2h_ed[j]) < 0)
									error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
								load_id_list = load_id_list -> next;
							}
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 5; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", w2h_fc[j]) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}
								
				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}

		if(hexas != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < hexas; i++){
				get_next_relevant_record(in_err_msg);
				if((output_type & 512) == 0){
					num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
											 &elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, &node6_id, &node7_id, &node8_id, 
											 &entity_type, &entity_id, &property_id,
											 &bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], &bound_ent_id[5], 
											 &bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4], &bound_ent_tp[5],
											 &bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4], &bound_ent_prop[5]);
					if(num != 24 && num != 30)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = (num == 30) ? YES : NO;
				}
				else{
					if(sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
										&elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, &node6_id, &node7_id, &node8_id, 
										&entity_type, &entity_id, &property_id,
										&bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], &bound_ent_id[5], 
										&bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4], &bound_ent_tp[5],
										&bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4], &bound_ent_prop[5],
										&bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], &bound_cur_id[3],
										&bound_cur_id[4], &bound_cur_id[5], &bound_cur_id[6], &bound_cur_id[7],
										&bound_cur_id[8], &bound_cur_id[9], &bound_cur_id[10], &bound_cur_id[11],
										&bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2], &bound_cur_prop[3],
										&bound_cur_prop[4], &bound_cur_prop[5], &bound_cur_prop[6], &bound_cur_prop[7],
										&bound_cur_prop[8], &bound_cur_prop[9], &bound_cur_prop[10], &bound_cur_prop[11]) != 54)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = YES;
				}
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != REGION_ENTITY)
						error_message("Invalid hexahedron entity detected", GENERAL_ERROR);
					
					entity = get_entity_id(region_list, entity_id);
					
/* the following should not be relevant */
					
					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){

#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif

						process_entity(entity, entity_type);
					}

					elem_type = entity -> hexa_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
						
					if(elem_type == NULL){
						sprintf(err_msg, "Element type specification for hexas in %s %ld missing", keyword[REGION], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
					
				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 8 %ld %ld %ld %ld %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, node4_id, node5_id, node6_id, node7_id, node8_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 6; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}

				if((output_type & 512) != 0){

					/* check whether bload has been processed on particular edge */

					edge_flag[0] = get_linear_edge_flag(node1_id, node2_id, bound_cur_id[0]);
					edge_flag[1] = get_linear_edge_flag(node2_id, node3_id, bound_cur_id[1]);
					edge_flag[2] = get_linear_edge_flag(node3_id, node4_id, bound_cur_id[2]);
					edge_flag[3] = get_linear_edge_flag(node4_id, node1_id, bound_cur_id[3]);
					edge_flag[4] = get_linear_edge_flag(node5_id, node6_id, bound_cur_id[4]);
					edge_flag[5] = get_linear_edge_flag(node6_id, node7_id, bound_cur_id[5]);
					edge_flag[6] = get_linear_edge_flag(node7_id, node8_id, bound_cur_id[6]);
					edge_flag[7] = get_linear_edge_flag(node8_id, node5_id, bound_cur_id[7]);
					edge_flag[8] = get_linear_edge_flag(node1_id, node5_id, bound_cur_id[8]);
					edge_flag[9] = get_linear_edge_flag(node2_id, node6_id, bound_cur_id[9]);
					edge_flag[10] = get_linear_edge_flag(node3_id, node7_id, bound_cur_id[10]);
					edge_flag[11] = get_linear_edge_flag(node4_id, node8_id, bound_cur_id[11]);

					for(j = 0; j < 12; j++){
						cflag[j] = NO;
						if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
							cflag[j] = YES;
							if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
								if(ent -> bload_id_list != NULL)
									count += get_list_size(ent -> bload_id_list);
								else
									cflag[j] = NO;
							}
							else
								cflag[j] = NO;
						}
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 6; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)   /* h2h_fc[j] */
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}

					if((output_type & 512) != 0){
						for(j = 0; j < 12; j++){
							if(cflag[j] == NO)continue;
							ent = get_entity_id(curve_list, bound_cur_id[j]);
							load_id_list = ent -> bload_id_list;
							while(load_id_list != NULL){
								if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), h2h_ed[j]) < 0)
									error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
								load_id_list = load_id_list -> next;
							}
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 6; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 6; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", j + 1) < 0)       /* h2h_fc[j] */
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 6; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}
								
				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}
	}
	
	if(elem_degree == QUADRATIC){
		if(edges != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < edges; i++){
				get_next_relevant_record(in_err_msg);
				if(sscanf(line_buffer, "%ld %ld %ld %ld %d %ld %ld", 
									&elem_id, &node1_id, &node2_id, &nd1_id, &entity_type, &entity_id, &property_id) != 7)
					error_message(in_err_msg, FILE_READ_ERROR);
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != CURVE_ENTITY)
						error_message("Invalid edge entity detected", GENERAL_ERROR);
					
					entity = get_entity_id(curve_list, entity_id);

/* the following should not be relevant */

					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){
						
#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif
						
						process_entity(entity, entity_type);
					}

					elem_type = entity -> edge_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
 					if(elem_type == NULL){
						sprintf(err_msg, "Element type specification for edges on %s %ld missing", keyword[CURVE], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
				
				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 3 %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, nd1_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}

				if((load_id_list = entity -> bload_id_list) != NULL){
					count = get_list_size(load_id_list);

					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					while(load_id_list != NULL){
						if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), 1) < 0)
							error_message(out_err_msg, FILE_WRITE_ERROR);      /* OOFEM elem record */
						load_id_list = load_id_list -> next;
					}
				}

				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}
		
		if(trias != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < trias; i++){
				get_next_relevant_record(in_err_msg);
				num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld", 
										 &elem_id, &node1_id, &node2_id, &node3_id, &nd1_id, &nd2_id, &nd3_id, 
										 &entity_type, &entity_id, &property_id,
										 &bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], 
										 &bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2]);
				if(num != 13 && num != 16)
					error_message(in_err_msg, FILE_READ_ERROR);
				bnd_prop = (num == 16) ? YES : NO;
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != SURFACE_ENTITY && entity_type != PATCH_ENTITY && entity_type != SHELL_ENTITY)
						error_message("Invalid triangle entity detected", GENERAL_ERROR);
					
					if(entity_type == SURFACE_ENTITY)entity = get_entity_id(surface_list, entity_id);
					if(entity_type == PATCH_ENTITY)entity = get_entity_id(patch_list, entity_id);
					if(entity_type == SHELL_ENTITY)entity = get_entity_id(shell_list, entity_id);
						
/* the following should not be relevant */

					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){
						
#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif
						
						process_entity(entity, entity_type);
					}

					elem_type = entity -> tria_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
 					if(elem_type == NULL){
						if(entity_type == SURFACE_ENTITY)key = SURFACE;
						if(entity_type == PATCH_ENTITY)key = PATCH;
						if(entity_type == SHELL_ENTITY)key = SHELL;
						sprintf(err_msg, "Element type specification for trias on %s %ld missing", keyword[key], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
				
				/* check whether bload and bcode has been processed on particular edge */
				
				edge_flag[0] = get_quadratic_edge_flag(nd1_id);
				edge_flag[1] = get_quadratic_edge_flag(nd2_id);
				edge_flag[2] = get_quadratic_edge_flag(nd3_id);

				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 6 %ld %ld %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, nd1_id, nd2_id, nd3_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 3; j++){
					bflag[j] = NO;
					if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
						bflag[j] = YES;
						if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 3; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}
				}

				count = 0;
				for(j = 0; j < 3; j++){
					bflag[j] = NO;
					if(bound_cur_id[j] != 0){
						bflag[j] = YES;
						if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 3; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 3; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}
				
				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}

		if(quads != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < quads; i++){
				get_next_relevant_record(in_err_msg);
				num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
										 &elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &nd1_id, &nd2_id, &nd3_id, &nd4_id,
										 &entity_type, &entity_id, &property_id,
										 &bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], &bound_cur_id[3],
										 &bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2], &bound_cur_prop[3]);
				if(num != 16 && num != 20)
					error_message(in_err_msg, FILE_READ_ERROR);
				bnd_prop = (num == 20) ? YES : NO;
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != SURFACE_ENTITY && entity_type != PATCH_ENTITY && entity_type != SHELL_ENTITY)
						error_message("Invalid quarilateral entity detected", GENERAL_ERROR);
					
					if(entity_type == SURFACE_ENTITY)entity = get_entity_id(surface_list, entity_id);
					if(entity_type == PATCH_ENTITY)entity = get_entity_id(patch_list, entity_id);
					if(entity_type == SHELL_ENTITY)entity = get_entity_id(shell_list, entity_id);
						
/* the following should not be relevant */

					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){
						
#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif
						
						process_entity(entity, entity_type);
					}

					elem_type = entity -> quad_type;
					
 					if(elem_type == NULL){
						if(entity_type == SURFACE_ENTITY)key = SURFACE;
						if(entity_type == PATCH_ENTITY)key = PATCH;
						if(entity_type == SHELL_ENTITY)key = SHELL;
						sprintf(err_msg, "Element type specification for quads on %s %ld missing", keyword[key], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
				
				/* check whether bload and bcode has been processed on particular edge */

				edge_flag[0] = get_quadratic_edge_flag(nd1_id);
				edge_flag[1] = get_quadratic_edge_flag(nd2_id);
				edge_flag[2] = get_quadratic_edge_flag(nd3_id);
				edge_flag[3] = get_quadratic_edge_flag(nd4_id);

				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 8 %ld %ld %ld %ld %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, node4_id, nd1_id, nd2_id, nd3_id, nd4_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 4; j++){
					bflag[j] = NO;
					if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
						bflag[j] = YES;
						if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}
				}

				count = 0;
				for(j = 0; j < 4; j++){
					bflag[j] = NO;
					if(bound_cur_id[j] != 0){
						bflag[j] = YES;
						if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						ent = get_entity_id(curve_list, bound_cur_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}
				
				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}
			
		if(tetras != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < tetras; i++){
				get_next_relevant_record(in_err_msg);
				if((output_type & 512) == 0){
					num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
											 &elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &nd1_id, &nd2_id, 
											 &nd3_id, &nd4_id, &nd5_id, &nd6_id, &entity_type, &entity_id, &property_id,
											 &bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], 
											 &bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3],
											 &bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3]);
					if(num != 22 && num != 26)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = (num == 26) ? YES : NO;
				}
				else{
					if(sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
										&elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &nd1_id, &nd2_id, 
										&nd3_id, &nd4_id, &nd5_id, &nd6_id, &entity_type, &entity_id, &property_id,
										&bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], 
										&bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3],
										&bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], 
										&bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], &bound_cur_id[3], &bound_cur_id[4], &bound_cur_id[5], 
										&bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2], 
										&bound_cur_prop[3], &bound_cur_prop[4], &bound_cur_prop[5]) != 38)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = YES;
				}
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != REGION_ENTITY)
						error_message("Invalid tetrahedron entity detected", GENERAL_ERROR);
					
					entity = get_entity_id(region_list, entity_id);
					
/* the following should not be relevant */

					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){
						
#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif
						
						process_entity(entity, entity_type);
					}

					elem_type = entity -> tetra_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
 					if(elem_type == NULL){
						sprintf(err_msg, "Element type specification for tetras in %s %ld missing", keyword[REGION], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
					
				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 10 %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, node4_id, 
									 nd1_id, nd2_id, nd3_id, nd4_id, nd5_id, nd6_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
					
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
						
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
					
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 4; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}

				if((output_type & 512) != 0){

					/* check whether bload has been processed on particular edge */

					edge_flag[0] = get_quadratic_edge_flag(nd1_id);
					edge_flag[1] = get_quadratic_edge_flag(nd2_id);
					edge_flag[2] = get_quadratic_edge_flag(nd3_id);
					edge_flag[3] = get_quadratic_edge_flag(nd4_id);
					edge_flag[4] = get_quadratic_edge_flag(nd5_id);
					edge_flag[5] = get_quadratic_edge_flag(nd6_id);

					for(j = 0; j < 6; j++){
						cflag[j] = NO;
						if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
							cflag[j] = YES;
							if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
								if(ent -> bload_id_list != NULL)
									count += get_list_size(ent -> bload_id_list);
								else
									cflag[j] = NO;
							}
							else
								cflag[j] = NO;
						}
					}
				}					
				
				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}

					if((output_type & 512) != 0){
						for(j = 0; j < 6; j++){
							if(cflag[j] == NO)continue;
							ent = get_entity_id(curve_list, bound_cur_id[j]);
							load_id_list = ent -> bload_id_list;
							while(load_id_list != NULL){
								if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)
									error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
								load_id_list = load_id_list -> next;
							}
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 4; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", j + 1) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 4; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}

				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}

		if(pyrams != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < pyrams; i++){
				get_next_relevant_record(in_err_msg);
				if((output_type & 512) == 0){
					num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
											 &elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, &nd1_id, &nd2_id, 
											 &nd3_id, &nd4_id, &nd5_id, &nd6_id, &nd7_id, &nd8_id,	&entity_type, &entity_id, &property_id,
											 &bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], 
											 &bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4],
											 &bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4]);
					if(num != 27 && num != 32)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = (num == 32) ? YES : NO;
				}
				else{
					if(sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
										&elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, &nd1_id, &nd2_id, 
										&nd3_id, &nd4_id, &nd5_id, &nd6_id, &nd7_id, &nd8_id,	&entity_type, &entity_id, &property_id,
										&bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], 
										&bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4],
										&bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4],
										&bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], &bound_cur_id[3],
										&bound_cur_id[4], &bound_cur_id[5], &bound_cur_id[6], &bound_cur_id[7],
										&bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2], &bound_cur_prop[3],
										&bound_cur_prop[4], &bound_cur_prop[5], &bound_cur_prop[6], &bound_cur_prop[7]) != 48)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = YES;
				}
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != REGION_ENTITY)
						error_message("Invalid pyramid entity detected", GENERAL_ERROR);
					
					entity = get_entity_id(region_list, entity_id);
					
/* the following should not be relevant */
					
					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){

#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif

						process_entity(entity, entity_type);
					}

					elem_type = entity -> pyram_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
					if(elem_type == NULL){
						sprintf(err_msg, "Element type specification for pyrams in %s %ld missing", keyword[REGION], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
					
				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 20 %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, node4_id, node5_id, node5_id, node5_id, node5_id, 
									 nd1_id, nd2_id, nd3_id, nd4_id, node5_id, node5_id, node5_id, node5_id, nd5_id, nd6_id, nd7_id, nd8_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 5; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if((output_type & 512) != 0){

					/* check whether bload has been processed on particular edge */

					edge_flag[0] = get_quadratic_edge_flag(nd1_id);
					edge_flag[1] = get_quadratic_edge_flag(nd2_id);
					edge_flag[2] = get_quadratic_edge_flag(nd3_id);
					edge_flag[3] = get_quadratic_edge_flag(nd4_id);
					edge_flag[4] = get_quadratic_edge_flag(nd5_id);
					edge_flag[5] = get_quadratic_edge_flag(nd6_id);
					edge_flag[6] = get_quadratic_edge_flag(nd7_id);
					edge_flag[7] = get_quadratic_edge_flag(nd8_id);
					
					for(j = 0; j < 8; j++){
						cflag[j] = NO;
						if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
							cflag[j] = YES;
							if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
								if(ent -> bload_id_list != NULL)
									count += get_list_size(ent -> bload_id_list);
								else
									cflag[j] = NO;
							}
							else
								cflag[j] = NO;
						}
					}
				}

				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), p2h_fc[j]) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}

					if((output_type & 512) != 0){
						for(j = 0; j < 8; j++){
							if(cflag[j] == NO)continue;
							ent = get_entity_id(curve_list, bound_cur_id[j]);
							load_id_list = ent -> bload_id_list;
							while(load_id_list != NULL){
								if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)   /* p2h_ed[j] */
									error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
								load_id_list = load_id_list -> next;
							}
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 5; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", p2h_fc[j]) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}
								
				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}

		if(wedges != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < wedges; i++){
				get_next_relevant_record(in_err_msg);
				if((output_type & 512) == 0){
					num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
											 &elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, &node6_id, 
											 &nd1_id, &nd2_id, &nd3_id, &nd4_id, &nd5_id, &nd6_id, &nd7_id, &nd8_id, &nd9_id, 
											 &entity_type, &entity_id, &property_id,
											 &bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], 
											 &bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4],
											 &bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4]);
					if(num != 29 && num != 34)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = (num == 34) ? YES : NO;
				}
				else{
					if(sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
										&elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, &node6_id, 
										&nd1_id, &nd2_id, &nd3_id, &nd4_id, &nd5_id, &nd6_id, &nd7_id, &nd8_id, &nd9_id, 
										&entity_type, &entity_id, &property_id,
										&bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], 
										&bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4],
										&bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4],
										&bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], &bound_cur_id[3], &bound_cur_id[4], 
										&bound_cur_id[5], &bound_cur_id[6], &bound_cur_id[7], &bound_cur_id[8],
										&bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2], &bound_cur_prop[3], &bound_cur_prop[4], 
										&bound_cur_prop[5], &bound_cur_prop[6], &bound_cur_prop[7], &bound_cur_prop[8]) != 52)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = YES;
				}
					
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != REGION_ENTITY)
						error_message("Invalid wedge entity detected", GENERAL_ERROR);
					
					entity = get_entity_id(region_list, entity_id);
					
/* the following should not be relevant */
					
					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){

#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif

						process_entity(entity, entity_type);
					}

					elem_type = entity -> wedge_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
					if(elem_type == NULL){
						sprintf(err_msg, "Element type specification for wedges in %s %ld missing", keyword[REGION], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
					
				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 20 %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, node3_id, node4_id, node5_id, node6_id, node6_id, 
									 nd1_id, nd2_id, node3_id, nd3_id, nd4_id, nd5_id, node6_id, nd6_id, nd7_id, nd8_id, nd9_id, nd9_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
				
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
					
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
				
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 5; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if((output_type & 512) != 0){

					/* check whether bload has been processed on particular edge */
					
					edge_flag[0] = get_quadratic_edge_flag(nd1_id);
					edge_flag[1] = get_quadratic_edge_flag(nd2_id);
					edge_flag[2] = get_quadratic_edge_flag(nd3_id);
					edge_flag[3] = get_quadratic_edge_flag(nd4_id);
					edge_flag[4] = get_quadratic_edge_flag(nd5_id);
					edge_flag[5] = get_quadratic_edge_flag(nd6_id);
					edge_flag[6] = get_quadratic_edge_flag(nd7_id);
					edge_flag[7] = get_quadratic_edge_flag(nd8_id);
					edge_flag[8] = get_quadratic_edge_flag(nd9_id);

					for(j = 0; j < 9; j++){
						cflag[j] = NO;
						if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
							cflag[j] = YES;
							if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
								if(ent -> bload_id_list != NULL)
									count += get_list_size(ent -> bload_id_list);
								else
									cflag[j] = NO;
							}
							else
								cflag[j] = NO;
						}
					}
				}

				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), w2h_fc[j]) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}

					if((output_type & 512) != 0){
						for(j = 0; j < 9; j++){
							if(cflag[j] == NO)continue;
							ent = get_entity_id(curve_list, bound_cur_id[j]);
							load_id_list = ent -> bload_id_list;
							while(load_id_list != NULL){
								if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), w2h_ed[j]) < 0)
									error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
								load_id_list = load_id_list -> next;
							}
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 5; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", w2h_fc[j]) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 5; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}
								
				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}

		if(hexas != 0){
			last_entity_type = 0;
			last_entity_id = 0;
			last_property_id = 0;
			for(i = 0; i < hexas; i++){
				get_next_relevant_record(in_err_msg);
				if((output_type & 512) == 0){
					num = sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
											 &elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, &node6_id, &node7_id, &node8_id, 
											 &nd1_id, &nd2_id, &nd3_id, &nd4_id, &nd5_id, &nd6_id, &nd7_id, &nd8_id, &nd9_id, &nd10_id, &nd11_id, &nd12_id, 
											 &entity_type, &entity_id, &property_id,
											 &bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], &bound_ent_id[5], 
											 &bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4], &bound_ent_tp[5],
											 &bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4], &bound_ent_prop[5]);
					if(num != 36 && num != 42)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = (num == 42) ? YES : NO;
				}
				else{
					if(sscanf(line_buffer, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %d %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
										&elem_id, &node1_id, &node2_id, &node3_id, &node4_id, &node5_id, &node6_id, &node7_id, &node8_id, 
										&nd1_id, &nd2_id, &nd3_id, &nd4_id, &nd5_id, &nd6_id, &nd7_id, &nd8_id, &nd9_id, &nd10_id, &nd11_id, &nd12_id, 
										&entity_type, &entity_id, &property_id,
										&bound_ent_id[0], &bound_ent_id[1], &bound_ent_id[2], &bound_ent_id[3], &bound_ent_id[4], &bound_ent_id[5], 
										&bound_ent_tp[0], &bound_ent_tp[1], &bound_ent_tp[2], &bound_ent_tp[3], &bound_ent_tp[4], &bound_ent_tp[5],
										&bound_ent_prop[0], &bound_ent_prop[1], &bound_ent_prop[2], &bound_ent_prop[3], &bound_ent_prop[4], &bound_ent_prop[5],
										&bound_cur_id[0], &bound_cur_id[1], &bound_cur_id[2], &bound_cur_id[3],
										&bound_cur_id[4], &bound_cur_id[5], &bound_cur_id[6], &bound_cur_id[7],
										&bound_cur_id[8], &bound_cur_id[9], &bound_cur_id[10], &bound_cur_id[11],
										&bound_cur_prop[0], &bound_cur_prop[1], &bound_cur_prop[2], &bound_cur_prop[3],
										&bound_cur_prop[4], &bound_cur_prop[5], &bound_cur_prop[6], &bound_cur_prop[7],
										&bound_cur_prop[8], &bound_cur_prop[9], &bound_cur_prop[10], &bound_cur_prop[11]) != 66)
						error_message(in_err_msg, FILE_READ_ERROR);
					bnd_prop = YES;
				}
				
				if(last_entity_type != entity_type || last_entity_id != entity_id){
					if(entity_type != REGION_ENTITY)
						error_message("Invalid hexahedron entity detected", GENERAL_ERROR);
					
					entity = get_entity_id(region_list, entity_id);
					
/* the following should not be relevant */

					if(entity == NULL)entity = create_missing_entity_id(entity_id, entity_type);
					if(entity -> selected == NO){
						
#ifdef DEFAULT_FIRST
						entity -> elem_prop_list = reverse_list(entity -> elem_prop_list);
#endif
						
						process_entity(entity, entity_type);
					}

					elem_type = entity -> hexa_type;
					if(elem_type == NULL)elem_type = entity -> elem_type;       /* for compatibility reasons */
					
 					if(elem_type == NULL){
						sprintf(err_msg, "Element type specification for hexas in %s %ld missing", keyword[REGION], entity_id);
						error_message(err_msg, GENERAL_ERROR);
					}
					last_entity_type = entity_type;
					last_entity_id = entity_id;
					entity -> on_output = YES;
				}
					
				entity -> elems++;
				if(fprintf(oofem_in_file, "%s %ld nodes 20 %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld", 
									 elem_type, elem_id, node1_id, node2_id, node3_id, node4_id, node5_id, node6_id, node7_id, node8_id, 
									 nd1_id, nd2_id, nd3_id, nd4_id, nd5_id, nd6_id, nd7_id, nd8_id, nd9_id, nd10_id, nd11_id, nd12_id) < 0)
					error_message(out_err_msg, FILE_WRITE_ERROR);          /* OOFEM elem record */
					
				elem_list = entity -> elem_prop_list;
				while(elem_list != NULL){
					string = (char *)(elem_list -> item);
					elem_list = elem_list -> next;
						
					if(fprintf(oofem_in_file, "%s", string) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
				}
					
				if(property_id > 0){
					if(last_property_id != property_id){
						ent = get_entity_id(property_list, property_id);
						last_property_id = property_id;
					}
					if(ent != NULL){
						elem_list = ent -> elem_prop_list;
						
						while(elem_list != NULL){
							string = (char *)(elem_list -> item);
							elem_list = elem_list -> next;
							
							if(fprintf(oofem_in_file, "%s", string) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 6; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bload_id_list != NULL)
								count += get_list_size(ent -> bload_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if((output_type & 512) != 0){

					/* check whether bload has been processed on particular edge */

					edge_flag[0] = get_quadratic_edge_flag(nd1_id);
					edge_flag[1] = get_quadratic_edge_flag(nd2_id);
					edge_flag[2] = get_quadratic_edge_flag(nd3_id);
					edge_flag[3] = get_quadratic_edge_flag(nd4_id);
					edge_flag[4] = get_quadratic_edge_flag(nd5_id);
					edge_flag[5] = get_quadratic_edge_flag(nd6_id);
					edge_flag[6] = get_quadratic_edge_flag(nd7_id);
					edge_flag[7] = get_quadratic_edge_flag(nd8_id);
					edge_flag[8] = get_quadratic_edge_flag(nd9_id);
					edge_flag[9] = get_quadratic_edge_flag(nd10_id);
					edge_flag[10] = get_quadratic_edge_flag(nd11_id);
					edge_flag[11] = get_quadratic_edge_flag(nd12_id);

					for(j = 0; j < 12; j++){
						cflag[j] = NO;
						if(bound_cur_id[j] != 0 && edge_flag[j] == NO){
							cflag[j] = YES;
							if((ent = get_entity_id(curve_list, bound_cur_id[j])) != NULL){
								if(ent -> bload_id_list != NULL)
									count += get_list_size(ent -> bload_id_list);
								else
									cflag[j] = NO;
							}
							else
								cflag[j] = NO;
						}
					}
				}

				if(count != 0){
					if(fprintf(oofem_in_file, " boundaryLoads %ld", count * 2) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */
					
					for(j = 0; j < 6; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						load_id_list = ent -> bload_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), j + 1) < 0)    /* h2h_fc[j] */
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							load_id_list = load_id_list -> next;
						}
					}

					if((output_type & 512) != 0){
						for(j = 0; j < 12; j++){
							if(cflag[j] == NO)continue;
							ent = get_entity_id(curve_list, bound_cur_id[j]);
							load_id_list = ent -> bload_id_list;
							while(load_id_list != NULL){
								if(fprintf(oofem_in_file, " %ld %d", *(long *)(load_id_list -> item), h2h_ed[j]) < 0)
									error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
								load_id_list = load_id_list -> next;
							}
						}
					}
				}
				
				count = 0;
				for(j = 0; j < 6; j++){
					bflag[j] = NO;
					if(bound_ent_id[j] != 0){
						bflag[j] = YES;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						if((ent = get_entity_id(entity_list, bound_ent_id[j])) != NULL){
							if(ent -> bcode_id_list != NULL)
								count += get_list_size(ent -> bcode_id_list);
							else
								bflag[j] = NO;
						}
						else
							bflag[j] = NO;
					}
				}
				
				if(count != 0){
					if(fprintf(oofem_in_file, " bsides %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 6; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(load_id_list != NULL){
							if(fprintf(oofem_in_file, " %d", j + 1) < 0)       /* h2h_fc[j] */
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}

					if(fprintf(oofem_in_file, " bcodes %ld", count) < 0)
						error_message(out_err_msg, FILE_WRITE_ERROR);        /* OOFEM elem record */

					for(j = 0; j < 6; j++){
						if(bflag[j] == NO)continue;
						switch(bound_ent_tp[j]){
						case SURFACE_ENTITY:
							entity_list = surface_list;
							break;
						case PATCH_ENTITY:
							entity_list = patch_list;
							break;
						case SHELL_ENTITY:
							entity_list = shell_list;
							break;
						default:
							error_message("Invalid entity type", GENERAL_ERROR);
						}
						ent = get_entity_id(entity_list, bound_ent_id[j]);
						code_id_list = ent -> bcode_id_list;
						while(code_id_list != NULL){
							if(fprintf(oofem_in_file, " %ld", *(long *)(code_id_list -> item)) < 0)
								error_message(out_err_msg, FILE_WRITE_ERROR);    /* OOFEM elem record */
							code_id_list = code_id_list -> next;
						}
					}
				}
								
				if(fprintf(oofem_in_file, "\n") < 0)error_message(out_err_msg, FILE_WRITE_ERROR);
			}
		}
	}

	rewind(active_in_file = control_file);
	sprintf(in_err_msg, "Unexpected end of control file %s", control_name);
	
/* skip output file, problem description, analysis, metastep, modules, initmodules, xfemman, domain, output manager 
	 and component size records */

	get_next_relevant_record(in_err_msg);
	get_next_relevant_record(in_err_msg);
	get_next_relevant_record(in_err_msg);
	for(i = 0; i < add_recs; i++)get_next_relevant_record(in_err_msg); /* OOFEM analysis additional records */
	get_next_relevant_record(in_err_msg);
	get_next_relevant_record(in_err_msg);
	get_next_relevant_record(in_err_msg);
	
	for(i = 0; i < csect; i++)
		read_write_until_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM cross section record */
	for(i = 0; i < mater; i++)
		read_write_until_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM material record */
	for(i = 0; i < bcond; i++)
		read_write_until_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM boundary condition record */
	for(i = 0; i < nic; i++)
		read_write_until_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM initial condition record */
	for(i = 0; i < timef; i++)
		read_write_until_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM time function record */
	
/* copy OOFEM check section (hopefully it is independent of t3d2merlin) */

	fprintf(oofem_in_file, "\n");

	write_rec = NO;
	while(get_next_record(NULL)){
		if(strncasecmp(line_buffer, "#%BEGIN_CHECK%", 14) == 0)write_rec = YES;
		if(write_rec == YES)write_current_record(out_err_msg);
		if(strncasecmp(line_buffer, "#%END_CHECK%", 12) == 0)break;
	}

	fclose(control_file);
	fclose(t3d_out_file);
	fclose(oofem_in_file);

	if(node_flag != NULL)free(node_flag);
	if(con_node1 != NULL)free(con_node1);
	if(con_node2 != NULL)free(con_node2);

	check_entity_list(vertex_list, keyword[VERTEX]);
	check_entity_list(curve_list, keyword[CURVE]);
	check_entity_list(surface_list, keyword[SURFACE]);
	check_entity_list(patch_list, keyword[PATCH]);
	check_entity_list(shell_list, keyword[SHELL]);
	check_entity_list(region_list, keyword[REGION]);
	
	destroy_entity_list(vertex_list);
	destroy_entity_list(curve_list);
	destroy_entity_list(surface_list);
	destroy_entity_list(patch_list);
	destroy_entity_list(shell_list);
	destroy_entity_list(region_list);

	destroy_entities_list(vertices_list);
	destroy_entities_list(curves_list);
	destroy_entities_list(surfaces_list);
	destroy_entities_list(patches_list);
	destroy_entities_list(shells_list);
	destroy_entities_list(regions_list);

	vertex_list = deep_destroy_list(vertex_list);
	curve_list = deep_destroy_list(curve_list);
	surface_list = deep_destroy_list(surface_list);
	patch_list = deep_destroy_list(patch_list);
	shell_list = deep_destroy_list(shell_list);
	region_list = deep_destroy_list(region_list);

	vertices_list = deep_destroy_list(vertices_list);
	curves_list = deep_destroy_list(curves_list);
	surfaces_list = deep_destroy_list(surfaces_list);
	patches_list = deep_destroy_list(patches_list);
	shells_list = deep_destroy_list(shells_list);
	regions_list = deep_destroy_list(regions_list);
	
	property_list = deep_destroy_list(property_list);
	
	string_list = deep_destroy_list(string_list);
	id_list = deep_destroy_list(id_list);
	
	if(domain == YES){
		sprintf(oofem_domain_name, "%s.domain", oofem_in_name);

		if((oofem_in_file = fopen(oofem_in_name, "r")) == NULL){
			sprintf(err_msg, "File %s opening error", oofem_in_name);
			error_message(err_msg, FILE_OPEN_ERROR);
		}
		if((oofem_domain_file = fopen(oofem_domain_name, "w")) == NULL){
			sprintf(err_msg, "File %s opening error", oofem_domain_name);
			error_message(err_msg, FILE_OPEN_ERROR);
		}

		active_in_file = oofem_in_file;
		active_out_file = oofem_domain_file;

		sprintf(in_err_msg, "Unexpected end of oofem in file %s", oofem_in_name);
		sprintf(out_err_msg, "File %s writing error", oofem_domain_name);

		get_next_relevant_record(in_err_msg);  /* OOFEM output file record */
		get_next_relevant_record(in_err_msg);  /* OOFEM problem description record */
		get_next_relevant_record(in_err_msg);  /* OOFEM analysis type record */

		for(i = 0; i < add_recs; i++)get_next_relevant_record(in_err_msg);  /* OOFEM analysis additional records */

		read_write_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM domain record */
		read_write_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM output manager record */

		read_write_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM number record */
		
		for(i = 0; i < nodes; i++)
			read_write_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM node record */
		for(i = 0; i < elems; i++)
			read_write_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM elem record */
		for(i = 0; i < csect; i++)
			read_write_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM csect record */
		for(i = 0; i < mater; i++)
			read_write_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM mater record */
		for(i = 0; i < bcond; i++)
			read_write_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM bcond record */
		for(i = 0; i < nic; i++)
			read_write_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM icond record */
		for(i = 0; i < timef; i++)
			read_write_next_relevant_record(in_err_msg, out_err_msg);  /* OOFEM timef record */

		fclose(oofem_in_file);
		fclose(oofem_domain_file);
	}

	if(warning == YES){
		fprintf(stderr, "\nThe above warnings need not necessarily be an error\n");
		fprintf(stderr, "(for example, there might be no internal node on an entity,\n");
		fprintf(stderr, "or elements on boundary entities are not in output by default)\n");
		fprintf(stderr, "but they are worth to be checked\n\n");
	}

	return(0);
}




static list_rec *
add_item_to_list_head(list_rec *any_list, void *item)
{
	list_rec *lst = NULL;
	
	lst = (list_rec *)malloc(sizeof(list_rec));
	if(lst == NULL)error_message("Memory allocation error", MEMORY_ERROR);
	
	lst -> item = item;
	lst -> next = any_list;
	
	return(lst);
}

static list_rec *
deep_destroy_list(list_rec *any_list)
{
	list_rec *lst = NULL;
	
	while(any_list != NULL){
		if(any_list -> item != NULL)free(any_list -> item);
		any_list = (lst = any_list) -> next;
		free(lst);
	}
	return(NULL);
}

static list_rec *
destroy_list(list_rec *any_list)
{
	list_rec *lst = NULL;
	
	while(any_list != NULL){
		any_list = (lst = any_list) -> next;
		free(lst);
	}
	return(NULL);
}

static list_rec *
reverse_list(list_rec *any_list)
{
	list_rec *head = NULL, *lst = NULL;

	if(any_list == NULL)return(NULL);
	if(any_list -> next == NULL)return(any_list);
	head = NULL;
	while((lst = any_list) != NULL){
		any_list = any_list -> next;
		lst -> next = head;
		head = lst;
	}
	return(head);
}

static long
get_list_size(list_rec *any_list)
{
	long count = 0L;

	while(any_list != NULL){
		any_list = any_list -> next;
		count++;
	}
	return(count);
}




static char *
get_next_record(char *err_msg)
{
	char *buf = NULL;

	if((buf = fgets(line_buffer, BUFFER_SIZE, active_in_file)) == NULL){
		if(!feof(active_in_file))error_message("Reading error", FILE_READ_ERROR);
		if(err_msg == NULL)return(NULL);
		error_message(err_msg, GENERAL_ERROR);
	}
	return(buf);
}

static char *
get_next_relevant_record(char *err_msg)
{
	char *buf = NULL, *character = NULL;
	
	do{
		if((buf = get_next_record(err_msg)) == NULL)return(NULL);
		character = line_buffer;
		while(*character == ' ')character++;
	}while(*character == '\n' || *character == REMARK);
	return(buf);
}


static char *
get_first_token(char *err_msg)
{
	char *token = NULL, *buf = NULL;

	while(YES){
		if((buf = get_next_relevant_record(err_msg)) == NULL)return(NULL);
		if((token = strtok(buf, SEPARATOR)) == NULL)continue;
		break;
	}
	return(token);
}

static char *
get_next_token(char *err_msg)
{
	char *token = NULL;

	if((token = strtok(NULL, SEPARATOR)) == NULL){
		if(err_msg == NULL)return(NULL);
		error_message(err_msg, FILE_READ_ERROR);
	}
	if(*token == REMARK)return(NULL);
	return(token);
}


static int
get_token_key(char *token, char keyword[1][KEY_SIZE], int keynum)
{
	char *t = token;
	int key;
	
	if(KEY_CASE == UPPER_CASE){
		while(*t != '\0'){
			*t = (char)toupper((int)*t);
			t++;
		}
	}
	if(KEY_CASE == LOWER_CASE){
		while(*t != '\0'){
			*t = (char)tolower((int)*t);
			t++;
		}
	}
	
/* do not use strcasecmp (it is not standard) */

	for(key = 0; key < keynum; key++){
		if(strcmp(token, keyword[key]) == 0)
			return(key);
	}
	return(keynum);
}


long
get_next_int(char *err_msg)
{
	return(get_int_from_token(get_next_token(err_msg)));
}

double
get_next_fpn(char *err_msg)
{
	return(get_fpn_from_token(get_next_token(err_msg)));
}


long
get_int_from_token(char *token)
{
	char *t_ptr = token, *t = token, err_msg[128];
	long int_num;

	int_num = strtol(token, &t_ptr, 10);
	if(token == t_ptr || *t_ptr != '\0'){
		sprintf(err_msg, "Integer number expected (\"%s\" - read)", t);
		error_message(err_msg, GENERAL_ERROR);
	}
	return(int_num);
}

double
get_fpn_from_token(char *token)
{
	char *t_ptr = token, *t = token, err_msg[128];
	double fpn_num;

	fpn_num = strtod(token, &t_ptr);
	if(token == t_ptr || *t_ptr != '\0'){
		sprintf(err_msg, "Floating point number expected (\"%s\" - read)", t);
		error_message(err_msg, GENERAL_ERROR);
	}
	return(fpn_num);
}


static void
write_current_record(char *err_msg)
{
	if(fprintf(active_out_file, "%s", line_buffer) < 0)
		error_message(err_msg, FILE_WRITE_ERROR);
}

static void
read_write_next_relevant_record(char *in_err_msg, char *out_err_msg)
{
	get_next_relevant_record(in_err_msg);
	write_current_record(out_err_msg);
}

static void
read_write_unrelevant_records(char *in_err_msg, char *out_err_msg)
{
	char *character = NULL;

	while(YES){
		get_next_record(in_err_msg);
		character = line_buffer;
		while(*character == ' ')character++;
		if(*character != '\n' && *character != REMARK)break;
		write_current_record(out_err_msg);
	}
}

static void
read_write_until_next_relevant_record(char *in_err_msg, char *out_err_msg)
{
	read_write_unrelevant_records(in_err_msg, out_err_msg);
	write_current_record(out_err_msg);
}



static long *
get_id(list_rec *id_list, long id)
{
	long *ident = NULL;

	while(id_list != NULL){
		ident = (long *)(id_list -> item);
		id_list = id_list -> next;
		if(*ident == id)return(ident);
	}
	return(NULL);
}

static long *
create_id(long id)
{
	long *ident = NULL;

	ident = (long *)malloc(sizeof(long));
	if(ident == NULL)error_message("Memory allocation error", MEMORY_ERROR);

	*ident = id;

	return(ident);
}


static entity_rec *
get_entity_id(list_rec *entity_list, long id)
{
	entity_rec *ent = NULL;

	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;
		if(ent -> id == id)return(ent);
	}
	return(NULL);
}

static entity_rec *
create_entity_id(long id)
{
	entity_rec *ent = NULL;

	ent = (entity_rec *)malloc(sizeof(entity_rec));
	if(ent == NULL)error_message("Memory allocation error", MEMORY_ERROR);

	ent -> id = id;
	ent -> selected = NO;
	ent -> node_prop_list = NULL;
	ent -> elem_prop_list = NULL;
	ent -> elem_type = NULL;
	ent -> edge_type = NULL;
	ent -> tria_type = NULL;
	ent -> quad_type = NULL;
	ent -> tetra_type = NULL;
	ent -> pyram_type = NULL;
	ent -> wedge_type = NULL;
	ent -> hexa_type = NULL;
	ent -> bload_id_list = NULL;
	ent -> bcode_id_list = NULL;
	ent -> node_type = NULL;
	ent -> elems = 0;
	ent -> on_output = NO;
	ent -> curve_id = NULL;

	return(ent);
}


static entities_rec *
create_entities(void)
{
	entities_rec *ents = NULL;

	ents = (entities_rec *)malloc(sizeof(entities_rec));
	if(ents == NULL)error_message("Memory allocation error", MEMORY_ERROR);

	ents -> id = NULL;
	ents -> selected = NO;
	ents -> node_prop_list = NULL;
	ents -> elem_prop_list = NULL;
	ents -> elem_type = NULL;
	ents -> edge_type = NULL;
	ents -> tria_type = NULL;
	ents -> quad_type = NULL;
	ents -> tetra_type = NULL;
	ents -> pyram_type = NULL;
	ents -> wedge_type = NULL;
	ents -> hexa_type = NULL;
	ents -> bload_id_list = NULL;
	ents -> bcode_id_list = NULL;
	ents -> node_type = NULL;

	return(ents);
}


static entity_rec *
create_missing_entity_id(long id, int entity_type)
{
	entity_rec *ent = NULL;

	ent = create_entity_id(id);
	switch(entity_type){
	case VERTEX_ENTITY:
		vertex_list = add_item_to_list_head(vertex_list, ent);
		break;
	case CURVE_ENTITY:
		curve_list = add_item_to_list_head(curve_list, ent);
		break;
	case SURFACE_ENTITY:
		surface_list = add_item_to_list_head(surface_list, ent);
		break;
	case PATCH_ENTITY:
		patch_list = add_item_to_list_head(patch_list, ent);
		break;
	case SHELL_ENTITY:
		shell_list = add_item_to_list_head(shell_list, ent);
		break;
	case REGION_ENTITY:
		region_list = add_item_to_list_head(region_list, ent);
		break;
	}
	return(ent);
}


static void
process_entity(entity_rec *ent, int entity_type)
{
	entities_rec *ents = NULL;
	list_rec *entities_list = NULL;
	list_rec *node_prop_list = NULL;
	list_rec *elem_prop_list = NULL;
	list_rec *load_id_list = NULL;
	list_rec *code_id_list = NULL;
	int key;
	char err_msg[128];

	if(ent -> selected == YES)return;

	switch(entity_type){
	case VERTEX_ENTITY:
		key = VERTEX;
		entities_list = vertices_list;
		break;
	case CURVE_ENTITY:
		key = CURVE;
		entities_list = curves_list;
		break;
	case SURFACE_ENTITY:
		key = SURFACE;
		entities_list = surfaces_list;
		break;
	case PATCH_ENTITY:
		key = PATCH;
		entities_list = patches_list;
		break;
	case SHELL_ENTITY:
		key = SHELL;
		entities_list = shells_list;
		break;
	case REGION_ENTITY:
		key = REGION;
		entities_list = regions_list;
		break;
	default:
		return;
	}
	
	while(entities_list != NULL){
		ents = (entities_rec *)(entities_list -> item);
		entities_list = entities_list -> next;

		if(get_id(ents -> id, ent -> id) != NULL)continue;
		
		node_prop_list = ents -> node_prop_list;
		while(node_prop_list != NULL){
			ent -> node_prop_list = add_item_to_list_head(ent -> node_prop_list, node_prop_list -> item);
			node_prop_list = node_prop_list -> next;
		}

		if(entity_type != VERTEX_ENTITY){
			elem_prop_list = ents -> elem_prop_list;
			while(elem_prop_list != NULL){
				ent -> elem_prop_list = add_item_to_list_head(ent -> elem_prop_list, elem_prop_list -> item);
				elem_prop_list = elem_prop_list -> next;
			}
			load_id_list = ents -> bload_id_list;
			while(load_id_list != NULL){
				ent -> bload_id_list = add_item_to_list_head(ent -> bload_id_list, load_id_list -> item);
				load_id_list = load_id_list -> next;
			}
			code_id_list = ents -> bcode_id_list;
			while(code_id_list != NULL){
				ent -> bcode_id_list = add_item_to_list_head(ent -> bcode_id_list, code_id_list -> item);
				code_id_list = code_id_list -> next;
			}

/* do not use strcasecmp (it is not standard) */
		
			if(ents -> elem_type != NULL && ent -> elem_type != NULL){
				if(strcmp(ents -> elem_type, ent -> elem_type) != 0){
					sprintf(err_msg, "Ambiguous element type specification for %s %ld", keyword[key], ent -> id);
					error_message(err_msg, GENERAL_ERROR);
				}
			}
			if(ent -> elem_type == NULL)ent -> elem_type = ents -> elem_type;

			switch(entity_type){
			case CURVE_ENTITY:
				if(ents -> edge_type != NULL && ent -> edge_type != NULL){
					if(strcmp(ents -> edge_type, ent -> edge_type) != 0){
						sprintf(err_msg, "Ambiguous edge type specification for %s %ld", keyword[key], ent -> id);
						error_message(err_msg, GENERAL_ERROR);
					}
				}
				if(ent -> edge_type == NULL)ent -> edge_type = ents -> edge_type;
				break;
			case SURFACE_ENTITY:
			case PATCH_ENTITY:
			case SHELL_ENTITY:
				if(ents -> tria_type != NULL && ent -> tria_type != NULL){
					if(strcmp(ents -> tria_type, ent -> tria_type) != 0){
						sprintf(err_msg, "Ambiguous tria type specification for %s %ld", keyword[key], ent -> id);
						error_message(err_msg, GENERAL_ERROR);
					}
				}
				if(ents -> quad_type != NULL && ent -> quad_type != NULL){
					if(strcmp(ents -> quad_type, ent -> quad_type) != 0){
						sprintf(err_msg, "Ambiguous quad type specification for %s %ld", keyword[key], ent -> id);
						error_message(err_msg, GENERAL_ERROR);
					}
				}
				if(ent -> tria_type == NULL)ent -> tria_type = ents -> tria_type;
				if(ent -> quad_type == NULL)ent -> quad_type = ents -> quad_type;
				break;
			case REGION_ENTITY:
				if(ents -> tetra_type != NULL && ent -> tetra_type != NULL){
					if(strcmp(ents -> tetra_type, ent -> tetra_type) != 0){
						sprintf(err_msg, "Ambiguous tetra type specification for %s %ld", keyword[key], ent -> id);
						error_message(err_msg, GENERAL_ERROR);
					}
				}
				if(ents -> hexa_type != NULL && ent -> hexa_type != NULL){
					if(strcmp(ents -> hexa_type, ent -> hexa_type) != 0){
						sprintf(err_msg, "Ambiguous hexa type specification for %s %ld", keyword[key], ent -> id);
						error_message(err_msg, GENERAL_ERROR);
					}
				}
				if(ent -> tetra_type == NULL)ent -> tetra_type = ents -> tetra_type;
				if(ent -> pyram_type == NULL)ent -> pyram_type = ents -> pyram_type;
				if(ent -> wedge_type == NULL)ent -> wedge_type = ents -> wedge_type;
				if(ent -> hexa_type == NULL)ent -> hexa_type = ents -> hexa_type;
				break;
			default:
				break;
			}

		}
	}

#ifdef DEFAULT_LAST
	ent -> node_prop_list = reverse_list(ent -> node_prop_list);
	ent -> node_prop_list = reverse_list(ent -> node_prop_list);
#endif

	ent -> selected = YES;
}
				

static void
process_entity_list(list_rec **entity_list, list_rec **entities_list, char *entity_name)
{
	entity_rec *ent = NULL;
	entities_rec *ents = NULL;
	char *token = NULL, err_msg[128];
	long id = -1;

	last_key = key;
	while((token = get_next_token(NULL)) != NULL){
		if((key = get_token_key(token, keyword, KEY_NUM)) == ALL){
			if(last_key != VERTEX && 
				 last_key != CURVE && 
				 last_key != SURFACE && 
				 last_key != PATCH && 
				 last_key != SHELL && 
				 last_key != REGION){
				sprintf(err_msg, "Unexpected keyword %s encountered", keyword[ALL]);
				error_message(err_msg, GENERAL_ERROR);
			}
			ents = create_entities();
			*entities_list = add_item_to_list_head(*entities_list, (void *)ents);
			ents -> selected = multi_select = YES;

			if((token = get_next_token(NULL)) != NULL){
				if((key = get_token_key(token, keyword, KEY_NUM)) != EXCEPT){
					sprintf(err_msg, "Invalid or unexpected keyword \"%s\" encountered", token);
					error_message(err_msg, GENERAL_ERROR);
				}
				process_entities_list(ents, entity_name);
			}
			return;
		}
			
		if((id = get_int_from_token(token)) <= 0){
			if(last_key != PROPERTY){
				sprintf(err_msg, "Invalid %s id (%ld)", entity_name, id);
				error_message(err_msg, GENERAL_ERROR);
			}
		}
		if((ent = get_entity_id(*entity_list, id)) == NULL){
			ent = create_entity_id(id);
			*entity_list = add_item_to_list_head(*entity_list, (void *)ent);
		}
		ent -> selected = YES;
	}

	if(id == -1){
		sprintf(err_msg, "Missing %s id", entity_name);
		error_message(err_msg, GENERAL_ERROR);
	}
}


static void
process_entities_list(entities_rec *ents, char *entity_name)
{
	char *token = NULL, err_msg[128];
	long id = 0, *ident = NULL;

	while((token = get_next_token(NULL)) != NULL){
		if((id = get_int_from_token(token)) <= 0){
			sprintf(err_msg, "Invalid %s id (%ld)", entity_name, id);
			error_message(err_msg, GENERAL_ERROR);
		}
		if((ident = get_id(id_list, id)) == NULL){
			ident = create_id(id);
			id_list = add_item_to_list_head(id_list, (void *)ident);
		}
		ents -> id = add_item_to_list_head(ents -> id, (void *)ident);
	}

	if(id == 0){
		sprintf(err_msg, "Missing %s id", entity_name);
		error_message(err_msg, GENERAL_ERROR);
	}
}


static void
deselect_entity_list(list_rec *entity_list)
{
	entity_rec *ent = NULL;

	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;
		ent -> selected = NO;
	}
}

static void
deselect_entities_list(list_rec *entities_list)
{
	entities_rec *ents = NULL;

	while(entities_list != NULL){
		ents = (entities_rec *)(entities_list -> item);
		entities_list = entities_list -> next;
		if(ents -> selected == NO)return;
		ents -> selected = NO;
	}
}


static void
reverse_node_elem_entity_list(list_rec *entity_list)
{
	entity_rec *ent = NULL;

	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;
		ent -> node_prop_list = reverse_list(ent -> node_prop_list);
		ent -> elem_prop_list = reverse_list(ent -> elem_prop_list);
	}
}


static void
apply_node_properties(char *string)
{
	if(entity_select == YES){
		if(vertex_select == YES)apply_node_properties_entity_list(vertex_list, string);
		if(curve_select == YES)apply_node_properties_entity_list(curve_list, string);
		if(surface_select == YES)apply_node_properties_entity_list(surface_list, string);
		if(patch_select == YES)apply_node_properties_entity_list(patch_list, string);
		if(shell_select == YES)apply_node_properties_entity_list(shell_list, string);
		if(region_select == YES)apply_node_properties_entity_list(region_list, string);
		
		if(multi_select == YES){
			if(vertex_select == YES)apply_node_properties_entities_list(vertices_list, string);
			if(curve_select == YES)apply_node_properties_entities_list(curves_list, string);
			if(surface_select == YES)apply_node_properties_entities_list(surfaces_list, string);
			if(patch_select == YES)apply_node_properties_entities_list(patches_list, string);
			if(shell_select == YES)apply_node_properties_entities_list(shells_list, string);
			if(region_select == YES)apply_node_properties_entities_list(regions_list, string);
		}
	}
	if(property_select == YES)apply_node_properties_entity_list(property_list, string);
}

static void
apply_node_properties_entity_list(list_rec *entity_list, char *string)
{
	entity_rec *ent = NULL;
	void *item = (void *)string;

	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;
		if(ent -> selected == NO)continue;
		ent -> node_prop_list = add_item_to_list_head(ent -> node_prop_list, item);
	}
}

static void
apply_node_properties_entities_list(list_rec *entities_list, char *string)
{
	entities_rec *ents = NULL;
	void *item = (void *)string;

	while(entities_list != NULL){
		ents = (entities_rec *)(entities_list -> item);
		entities_list = entities_list -> next;
		if(ents -> selected == NO)return;
		ents -> node_prop_list = add_item_to_list_head(ents -> node_prop_list, item);
	}
}


static void
apply_elem_properties(char *string)
{
	if(entity_select == YES){
/*	  if(vertex_select == YES)apply_elem_properties_entity_list(vertex_list, string); */
		if(curve_select == YES)apply_elem_properties_entity_list(curve_list, string);
		if(surface_select == YES)apply_elem_properties_entity_list(surface_list, string);
		if(patch_select == YES)apply_elem_properties_entity_list(patch_list, string);
		if(shell_select == YES)apply_elem_properties_entity_list(shell_list, string);
		if(region_select == YES)apply_elem_properties_entity_list(region_list, string);

		if(multi_select == YES){
/*			if(vertex_select == YES)apply_elem_properties_entities_list(vertices_list, string); */
			if(curve_select == YES)apply_elem_properties_entities_list(curves_list, string);
			if(surface_select == YES)apply_elem_properties_entities_list(surfaces_list, string);
			if(patch_select == YES)apply_elem_properties_entities_list(patches_list, string);
			if(shell_select == YES)apply_elem_properties_entities_list(shells_list, string);
			if(region_select == YES)apply_elem_properties_entities_list(regions_list, string);
		}
	}
	if(property_select == YES)apply_elem_properties_entity_list(property_list, string);
}

static void
apply_elem_properties_entity_list(list_rec *entity_list, char *string)
{
	entity_rec *ent = NULL;
	void *item = (void *)string;

	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;
		if(ent -> selected == NO)continue;
		ent -> elem_prop_list = add_item_to_list_head(ent -> elem_prop_list, item);
	}
}

static void
apply_elem_properties_entities_list(list_rec *entities_list, char *string)
{
	entities_rec *ents = NULL;
	void *item = (void *)string;

	while(entities_list != NULL){
		ents = (entities_rec *)(entities_list -> item);
		entities_list = entities_list -> next;
		if(ents -> selected == NO)return;
		ents -> elem_prop_list = add_item_to_list_head(ents -> elem_prop_list, item);
	}
}


static void
apply_node_type(char *string)
{
	if(entity_select == YES){
	  if(vertex_select == YES)apply_node_type_entity_list(vertex_list, string, keyword[VERTEX]);
		if(curve_select == YES)apply_node_type_entity_list(curve_list, string, keyword[CURVE]);
		if(surface_select == YES)apply_node_type_entity_list(surface_list, string, keyword[SURFACE]);
		if(patch_select == YES)apply_node_type_entity_list(patch_list, string, keyword[PATCH]);
		if(shell_select == YES)apply_node_type_entity_list(shell_list, string, keyword[SHELL]);
		if(region_select == YES)apply_node_type_entity_list(region_list, string, keyword[REGION]);

		if(multi_select == YES){
			if(vertex_select == YES)apply_node_type_entities_list(vertices_list, string, keyword[VERTEX]);
			if(curve_select == YES)apply_node_type_entities_list(curves_list, string, keyword[CURVE]);
			if(surface_select == YES)apply_node_type_entities_list(surfaces_list, string, keyword[SURFACE]);
			if(patch_select == YES)apply_node_type_entities_list(patches_list, string, keyword[PATCH]);
			if(shell_select == YES)apply_node_type_entities_list(shells_list, string, keyword[SHELL]);
			if(region_select == YES)apply_node_type_entities_list(regions_list, string, keyword[REGION]);
		}
	}
}

static void
apply_node_type_entity_list(list_rec *entity_list, char *string, char *entity_name)
{
	entity_rec *ent = NULL;
	char err_msg[128];

	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;
		if(ent -> selected == NO)continue;
		if(ent -> node_type != NULL){
			sprintf(err_msg, "Duplication of node type specification for %s %ld", entity_name, ent -> id);
			error_message(err_msg, GENERAL_ERROR);
		}
		ent -> node_type = string;
	}
}

static void
apply_node_type_entities_list(list_rec *entities_list, char *string, char *entity_name)
{
	entities_rec *ents = NULL;

	while(entities_list != NULL){
		ents = (entities_rec *)(entities_list -> item);
		entities_list = entities_list -> next;
		if(ents -> selected == NO)return;
		ents -> node_type = string;
	}
}



static void
apply_elem_type(char *string)
{
	if(entity_select == YES){
/*	  if(vertex_select == YES)apply_elem_type_entity_list(vertex_list, string, keyword[VERTEX]); */
		if(curve_select == YES)apply_elem_type_entity_list(curve_list, string, keyword[CURVE]);
		if(surface_select == YES)apply_elem_type_entity_list(surface_list, string, keyword[SURFACE]);
		if(patch_select == YES)apply_elem_type_entity_list(patch_list, string, keyword[PATCH]);
		if(shell_select == YES)apply_elem_type_entity_list(shell_list, string, keyword[SHELL]);
		if(region_select == YES)apply_elem_type_entity_list(region_list, string, keyword[REGION]);

		if(multi_select == YES){
/*			if(vertex_select == YES)apply_elem_type_entities_list(vertices_list, string, keyword[VERTEX]); */
			if(curve_select == YES)apply_elem_type_entities_list(curves_list, string, keyword[CURVE]);
			if(surface_select == YES)apply_elem_type_entities_list(surfaces_list, string, keyword[SURFACE]);
			if(patch_select == YES)apply_elem_type_entities_list(patches_list, string, keyword[PATCH]);
			if(shell_select == YES)apply_elem_type_entities_list(shells_list, string, keyword[SHELL]);
			if(region_select == YES)apply_elem_type_entities_list(regions_list, string, keyword[REGION]);
		}
	}
}

static void
apply_elem_type_entity_list(list_rec *entity_list, char *string, char *entity_name)
{
	entity_rec *ent = NULL;
	char err_msg[128];

	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;
		if(ent -> selected == NO)continue;
		switch(key){
		case ELEMTYPE:
			if(ent -> elem_type != NULL){
				sprintf(err_msg, "Duplication of element type specification for %s %ld", entity_name, ent -> id);
				error_message(err_msg, GENERAL_ERROR);
			}
			ent -> elem_type = string;
			break;
		case EDGETYPE:
			if(ent -> edge_type != NULL){
				sprintf(err_msg, "Duplication of edge type specification for %s %ld", entity_name, ent -> id);
				error_message(err_msg, GENERAL_ERROR);
			}
			ent -> edge_type = string;
			break;
		case TRIATYPE:
			if(ent -> tria_type != NULL){
				sprintf(err_msg, "Duplication of tria type specification for %s %ld", entity_name, ent -> id);
				error_message(err_msg, GENERAL_ERROR);
			}
			ent -> tria_type = string;
			break;
		case QUADTYPE:
			if(ent -> quad_type != NULL){
				sprintf(err_msg, "Duplication of quad type specification for %s %ld", entity_name, ent -> id);
				error_message(err_msg, GENERAL_ERROR);
			}
			ent -> quad_type = string;
			break;
		case TETRATYPE:
			if(ent -> tetra_type != NULL){
				sprintf(err_msg, "Duplication of tetra type specification for %s %ld", entity_name, ent -> id);
				error_message(err_msg, GENERAL_ERROR);
			}
			ent -> tetra_type = string;
			break;
		case HEXATYPE:
			if(ent -> hexa_type != NULL){
				sprintf(err_msg, "Duplication of hexa type specification for %s %ld", entity_name, ent -> id);
				error_message(err_msg, GENERAL_ERROR);
			}
			ent -> hexa_type = string;

/* pyrams and wedges are treated as degenerated hexas */

			ent -> pyram_type = string;
			ent -> wedge_type = string;
			break;
		default:
			break;
		}
	}
}

static void
apply_elem_type_entities_list(list_rec *entities_list, char *string, char *entity_name)
{
	entities_rec *ents = NULL;

	while(entities_list != NULL){
		ents = (entities_rec *)(entities_list -> item);
		entities_list = entities_list -> next;
		if(ents -> selected == NO)return;
		switch(key){
		case ELEMTYPE:
			ents -> elem_type = string;
			break;
		case EDGETYPE:
			ents -> edge_type = string;
			break;
		case TRIATYPE:
			ents -> tria_type = string;
			break;
		case QUADTYPE:
			ents -> quad_type = string;
			break;
		case TETRATYPE:
			ents -> tetra_type = string;
			break;
		case HEXATYPE:
			ents -> hexa_type = string;

/* pyrams and wedges are treated as degenerated hexas */

			ents -> pyram_type = string;
			ents -> wedge_type = string;
			break;
		default:
			break;
		}
	}
}


static void
apply_elem_bload(long *ident)
{
	if(entity_select == YES){
/*	  if(vertex_select == YES)apply_elem_bload_entity_list(vertex_list, ident); */
		if(curve_select == YES)apply_elem_bload_entity_list(curve_list, ident);
		if(surface_select == YES)apply_elem_bload_entity_list(surface_list, ident);
		if(patch_select == YES)apply_elem_bload_entity_list(patch_list, ident);
		if(shell_select == YES)apply_elem_bload_entity_list(shell_list, ident);
		if(region_select == YES)apply_elem_bload_entity_list(region_list, ident);

		if(multi_select == YES){
/*			if(vertex_select == YES)apply_elem_bload_entities_list(vertices_list, ident); */
			if(curve_select == YES)apply_elem_bload_entities_list(curves_list, ident);
			if(surface_select == YES)apply_elem_bload_entities_list(surfaces_list, ident);
			if(patch_select == YES)apply_elem_bload_entities_list(patches_list, ident);
			if(shell_select == YES)apply_elem_bload_entities_list(shells_list, ident);
			if(region_select == YES)apply_elem_bload_entities_list(regions_list, ident);
		}
	}
}

static void
apply_elem_bload_entity_list(list_rec *entity_list, long *ident)
{
	entity_rec *ent = NULL;
	void *item = (void *)ident;

	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;
		if(ent -> selected == NO)continue;
		ent -> bload_id_list = add_item_to_list_head(ent -> bload_id_list, item);
	}
}


static void
apply_elem_bload_entities_list(list_rec *entities_list, long *ident)
{
	entities_rec *ents = NULL;
	void *item = (void *)ident;

	while(entities_list != NULL){
		ents = (entities_rec *)(entities_list -> item);
		entities_list = entities_list -> next;
		if(ents -> selected == NO)continue;
		ents -> bload_id_list = add_item_to_list_head(ents -> bload_id_list, item);
	}
}




static void
apply_elem_bcode(long *ident)
{
	if(entity_select == YES){
/*	  if(vertex_select == YES)apply_elem_bcode_entity_list(vertex_list, ident); */
		if(curve_select == YES)apply_elem_bcode_entity_list(curve_list, ident);
		if(surface_select == YES)apply_elem_bcode_entity_list(surface_list, ident);
		if(patch_select == YES)apply_elem_bcode_entity_list(patch_list, ident);
		if(shell_select == YES)apply_elem_bcode_entity_list(shell_list, ident);
		if(region_select == YES)apply_elem_bcode_entity_list(region_list, ident);

		if(multi_select == YES){
/*			if(vertex_select == YES)apply_elem_bcode_entities_list(vertices_list, ident); */
			if(curve_select == YES)apply_elem_bcode_entities_list(curves_list, ident);
			if(surface_select == YES)apply_elem_bcode_entities_list(surfaces_list, ident);
			if(patch_select == YES)apply_elem_bcode_entities_list(patches_list, ident);
			if(shell_select == YES)apply_elem_bcode_entities_list(shells_list, ident);
			if(region_select == YES)apply_elem_bcode_entities_list(regions_list, ident);
		}
	}
}

static void
apply_elem_bcode_entity_list(list_rec *entity_list, long *ident)
{
	entity_rec *ent = NULL;
	void *item = (void *)ident;

	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;
		if(ent -> selected == NO)continue;
		ent -> bcode_id_list = add_item_to_list_head(ent -> bcode_id_list, item);
	}
}


static void
apply_elem_bcode_entities_list(list_rec *entities_list, long *ident)
{
	entities_rec *ents = NULL;
	void *item = (void *)ident;

	while(entities_list != NULL){
		ents = (entities_rec *)(entities_list -> item);
		entities_list = entities_list -> next;
		if(ents -> selected == NO)continue;
		ents -> bcode_id_list = add_item_to_list_head(ents -> bcode_id_list, item);
	}
}


/* note: in linear case, edge flag is processed only if there is boundary curve */

static logic
get_linear_edge_flag(long node1_id, long node2_id, long curve_id)
{
	entity_rec *ent1 = NULL, *ent2 = NULL;
	long nd1_id, nd2_id, nd_id, *ident = NULL;

	if(curve_id == 0)return(NO);

	nd1_id = con_node1[node1_id - 1];
	if(nd1_id == 0){                         /* node1 is not yet processed curve node */
		con_node1[node1_id - 1] = node2_id;

		/* update con nodes of node2 unless node2 is vertex */
		nd_id = con_node1[node2_id - 1];
		if(nd_id == 0)                        /* node2 is not yet processed curve node */
			con_node1[node2_id - 1] = node1_id;
		else{
			if(nd_id > 0){                      /* node2 is at least once processed curve node */
				if(nd_id != node1_id)con_node2[node2_id - 1] = node1_id;
			}
		}
		return(NO);
	}
	if(nd1_id > 0){                          /* node1 is at least once processed curve node */
		if(nd1_id == node2_id)return(YES);

		nd_id = con_node2[node1_id - 1];
		if(nd_id == 0){
			con_node2[node1_id - 1] = node2_id;

			/* update con nodes of node2 unless node2 is vertex */
			nd_id = con_node1[node2_id - 1];
			if(nd_id == 0)                      /* node2 is not yet processed curve node */
				con_node1[node2_id - 1] = node1_id;
			else{
				if(nd_id > 0){                    /* node2 is at least once processed curve node */
					if(nd_id != node1_id)con_node2[node2_id - 1] = node1_id;
				}
			}
			return(NO);
		}

		if(nd_id > 0){                        /* node2 is twice processed curve node */
			if(nd_id != node2_id)
				error_message("Unexpected situation", GENERAL_ERROR);
		}

		return(YES);
	}
			
	nd2_id = con_node1[node2_id - 1];
	if(nd2_id == 0){                         /* node2 is not yet processed curve node */
		con_node1[node2_id - 1] = node1_id;

		/* update con nodes of node1 unless node2 is vertex */
		nd_id = con_node1[node1_id - 1];
		if(nd_id == 0)                        /* node1 is not yet processed curve node */
			con_node1[node1_id - 1] = node2_id;
		else{
			if(nd_id > 0){                      /* node1 is at least once processed curve node */
				if(nd_id != node2_id)con_node2[node1_id - 1] = node2_id;
			}
		}
		return(NO);
	}
	if(nd2_id > 0){                          /* node2 is at least once processed curve node */
		if(nd2_id == node1_id)return(YES);

		nd_id = con_node2[node2_id - 1];
		if(nd_id == 0){
			con_node2[node2_id - 1] = node1_id;

			/* update con nodes of node1 unless node1 is vertex */
			nd_id = con_node1[node1_id - 1];
			if(nd_id == 0)                      /* node1 is not yet processed curve node */
				con_node1[node1_id - 1] = node2_id;
			else{
				if(nd_id > 0){                    /* node1 is at least once processed curve node */
					if(nd_id != node2_id)con_node2[node1_id - 1] = node2_id;
				}
			}
			return(NO);
		}

		if(nd_id > 0){                        /* node1 is twice processed curve node */
			if(nd_id != node1_id)
				error_message("Unexpected situation", GENERAL_ERROR);
		}

		return(YES);
	}

	/* node1 as well as node2 are vertices;
		 use list of ids of curves connected to the vertex;
		 these lists are available only if both nodes are vertices **/

	ent1 = get_entity_id(vertex_list, -nd1_id);
	ent2 = get_entity_id(vertex_list, -nd2_id);

	if((ident = get_id(ent1 -> curve_id, curve_id)) != NULL)return(YES);
	if((ident = get_id(ent2 -> curve_id, curve_id)) != NULL)error_message("Unexpected situation", GENERAL_ERROR);

	ident = create_id(curve_id);
	id_list = add_item_to_list_head(id_list, (void *)ident);
	ent1 -> curve_id = add_item_to_list_head(ent1 -> curve_id, (void *)ident);
	ent2 -> curve_id = add_item_to_list_head(ent2 -> curve_id, (void *)ident);

	return(NO);
}


/* note: in quadratic case, edge flag is processed irrespectably whether there is boundary curve */

static logic
get_quadratic_edge_flag(long midnode_id)
{
	if(node_flag[midnode_id - 1] == 0){
		node_flag[midnode_id - 1] = 1;
		return(NO);
	}
	return(YES);
}


static void
delete_entities_properties(list_rec *entities_list)
{
	entities_rec *ents = NULL;
	
	while(entities_list != NULL){
		ents = (entities_rec *)(entities_list -> item);
		entities_list = entities_list -> next;
	
		if(ents -> node_prop_list != NULL)ents -> node_prop_list = destroy_list(ents -> node_prop_list);
		if(ents -> elem_prop_list != NULL)ents -> elem_prop_list = destroy_list(ents -> elem_prop_list);
		if(ents -> bload_id_list != NULL)ents -> bload_id_list = destroy_list(ents -> bload_id_list);
		if(ents -> bcode_id_list != NULL)ents -> bcode_id_list = destroy_list(ents -> bcode_id_list);
	}
}



static void
destroy_entity_list(list_rec *entity_list)
{
	entity_rec *ent = NULL;
	
	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;
	
		if(ent -> node_prop_list != NULL)ent -> node_prop_list = destroy_list(ent -> node_prop_list);
		if(ent -> elem_prop_list != NULL)ent -> elem_prop_list = destroy_list(ent -> elem_prop_list);
		if(ent -> bload_id_list != NULL)ent -> bload_id_list = destroy_list(ent -> bload_id_list);
		if(ent -> bcode_id_list != NULL)ent -> bcode_id_list = destroy_list(ent -> bcode_id_list);
	}
}


static void
destroy_entities_list(list_rec *entities_list)
{
	entities_rec *ents = NULL;
	
	while(entities_list != NULL){
		ents = (entities_rec *)(entities_list -> item);
		entities_list = entities_list -> next;
	
		if(ents -> id != NULL)ents -> id = destroy_list(ents -> id);

		if(ents -> node_prop_list != NULL)ents -> node_prop_list = destroy_list(ents -> node_prop_list);
		if(ents -> elem_prop_list != NULL)ents -> elem_prop_list = destroy_list(ents -> elem_prop_list);
		if(ents -> bload_id_list != NULL)ents -> bload_id_list = destroy_list(ents -> bload_id_list);
		if(ents -> bcode_id_list != NULL)ents -> bcode_id_list = destroy_list(ents -> bcode_id_list);
	}
}



static void
check_entity_list(list_rec *entity_list, char *entity_name)
{
	entity_rec *ent = NULL;
	
	while(entity_list != NULL){
		ent = (entity_rec *)(entity_list -> item);
		entity_list = entity_list -> next;

		if(ent -> on_output == NO){
			fprintf(stderr, "Warning: %s %ld referenced in ctrl file is not present in t3d output file\n",
							entity_name, ent -> id);
			warning = YES;
		}
	}
}
			

static char *
store_string(char *string)
{
	char *str = NULL;

	str = (char *)malloc((int)strlen(string) + 1);
	if(str == NULL)error_message("Memory allocation error", MEMORY_ERROR);
	strcpy(str,string);
	return(str);
}


static void
exit_error_message(char *message, int exit_code, int line)
{
	fprintf(stderr, "\nt3d2oofem: line %d\n%s\n", line, message);
	if(exit_code == FILE_READ_ERROR){
		fprintf(stderr, "%s\n", line_buffer);
	}
	fprintf(stderr, "\n");
	exit(exit_code);
}
